import json
import sys
from pathlib import Path
from typing import Any, Type

from redb.core import BaseDocument, Document
from redb.interface.errors import DocumentNotFound
from redb.interface.collection import Collection, Json, OptionalJson, ReturnType
from redb.interface.fields import CompoundIndex, PyMongoOperations
from redb.interface.results import (
    BulkWriteResult,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)


class JSONCollection(Collection):
    __client_name__ = "json"

    def __init__(self, collection: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__collection = collection

    def _get_driver_collection(self):
        return self.__collection

    def create_index(
        self,
        index: CompoundIndex,
    ) -> bool:
        raise NotImplementedError

    def find(
        self,
        cls: Type[Document],
        return_cls: Type[ReturnType],
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        sort: list[tuple[str, str | int]] | None = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[ReturnType]:
        transform = lambda file_path: json.load(open(file_path))
        if filter is not None and "_id" in filter:
            file_path = self.__collection / f"{filter['_id']}.json"
            if file_path.is_file() and not file_path.is_symlink():
                return [return_cls(**transform(file_path))]

            if len(filter.keys()) == 1:
                # If the only filter was the ID, return empty list
                return []

        json_files = self.__collection.glob("*.json")
        out = []
        for i, json_file in enumerate(json_files):
            if i < skip:
                continue
            if limit and len(out) >= limit:
                break
            if not json_file.is_file():
                continue

            transformed_json: dict = transform(json_file)
            if filter is not None:
                ignore_file = False
                for key in filter:
                    if key not in transformed_json:
                        ignore_file = True
                        break
                    if transformed_json[key] != filter[key]:
                        ignore_file = True
                        break
                if ignore_file:
                    continue

            if fields is not None:
                transformed_json = {
                    key: value
                    for key, value in transformed_json.items()
                    if key in fields
                }

            out.append(return_cls(**transformed_json))

        return out

    def find_one(
        self,
        cls: Type[Document],
        return_cls: Type[ReturnType],
        fields: dict[str, bool] | None = None,
        filter: OptionalJson = None,
        skip: int = 0,
    ) -> ReturnType:
        captures = self.find(
            cls,
            return_cls,
            filter=filter,
            fields=fields,
            skip=skip,
            limit=1,
        )
        return captures[0]

    def distinct(
        self,
        cls: Type[ReturnType],
        key: str,
        filter: OptionalJson = None,
    ) -> list[Any]:
        docs = self.find(
            cls,
            return_cls=dict,
            filter=filter,
        )
        computed_values = set()
        for doc in docs:
            if doc[key] in computed_values:
                continue

            computed_values.add(doc[key])

        return list(computed_values)

    def count_documents(
        self,
        cls: Type[Document],
        filter: OptionalJson = None,
    ) -> int:
        return len(self.find(cls, return_cls=dict, filter=filter))

    def bulk_write(self, _: list[PyMongoOperations]) -> BulkWriteResult:
        raise NotImplementedError

    def insert_one(
        self,
        cls: Type[Document],
        data: Json,
    ) -> InsertOneResult:
        self.__collection.mkdir(parents=True, exist_ok=True)

        id = data["_id"]
        json_path = self.__collection / Path(f"{id}.json")
        if json_path.is_file():
            raise ValueError(f"Document with {id} already exists")

        with open(json_path, "w") as f:
            json.dump(data, f, indent=4)

        return InsertOneResult(inserted_id=id)

    def insert_many(
        self,
        cls: Type[Document],
        data: list[Json],
    ) -> InsertManyResult:
        ids = []
        for item in data:
            ids.append(self.insert_one(cls, data=item).inserted_id)

        return InsertManyResult(inserted_ids=ids)

    def replace_one(
        self,
        cls: Type[Document],
        filter: Json,
        replacement: Json,
        upsert: bool = False,
    ) -> ReplaceOneResult:
        doc = self.find_one(cls, return_cls=dict, filter=filter)
        if doc is None:
            if not upsert:
                raise ValueError(f"Document not found")

            self.insert_one(cls, data=replacement)
            return ReplaceOneResult(
                matched_count=1,
                modified_count=1,
                upserted_id=replacement["_id"],
            )

        original_path = self.__collection / Path(f"{doc['_id']}.json")
        upserted = False
        if replacement["_id"] != doc["_id"]:
            # Since the ID has changed, we need to remove the old one
            original_path.unlink()
            upserted = True

        new_path = self.__collection / Path(f"{replacement['_id']}.json")
        with open(new_path, "w") as f:
            json.dump(replacement, f, indent=4)

        return ReplaceOneResult(
            matched_count=1,
            modified_count=1,
            upserted_id=replacement["_id"] if upserted else None,
        )

    def update_one(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateOneResult:
        doc: dict = self.find_one(cls, return_cls=dict, filter=filter)
        update = update.pop(next(iter(update.keys())))  # { $set: {...} }
        if doc is None:
            if not upsert:
                raise DocumentNotFound
            result = self.insert_one(cls, data=update)
            return UpdateOneResult(
                matched_count=1,
                modified_count=1,
                upserted_id=result.inserted_id,
            )
        doc_path = self.__collection / Path(f"{doc['_id']}.json")
        updated_content = doc | update
        with open(doc_path, "w") as f:
            json.dump(updated_content, f, indent=2)
        return UpdateOneResult(
            matched_count=1,
            modified_count=1,
            upserted_id=None,
        )

    def update_many(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateManyResult:
        upserted_ids = []
        docs = self.find(cls, return_cls=dict, filter=filter)
        if not docs:
            if not upsert:
                raise ValueError(f"Document not found")

            self.insert_one(cls, data=update)
            return UpdateManyResult(
                matched_count=1,
                modified_count=1,
                upserted_id=update["_id"],
            )

        for doc in docs:
            original_path = self.__collection / Path(f"{doc['_id']}.json")
            with open(original_path, "r") as f:
                original_content: dict = json.load(f)

            original_content.update(update)
            if "_id" in update:
                new_id = update["_id"]
            elif cls is dict:
                new_id = BaseDocument().get_dict_hash(
                    data=original_content, use_data_fields=True
                )
            else:
                new_id = cls.get_hash(data=original_content)

            if doc["_id"] != new_id:
                # Since the ID has changed, we need to remove the old one
                original_path.unlink()
                upserted_ids.append(new_id)

            original_content["_id"] = new_id
            new_path = self.__collection / Path(f"{new_id}.json")
            with open(new_path, "w") as f:
                json.dump(original_content, f, indent=4)

        return UpdateManyResult(
            matched_count=len(filter),
            modified_count=len(upserted_ids),
            upserted_id=upserted_ids,
        )

    def delete_one(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteOneResult:
        doc = self.find_one(cls, return_cls=dict, filter=filter)
        if doc is None:
            raise ValueError(f"Document not found")

        file = self.__collection / f"{doc['_id']}.json"
        file.unlink()
        return DeleteOneResult(deleted_count=1)

    def delete_many(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteManyResult:
        docs = self.find(cls, return_cls=dict, filter=filter)
        for doc in docs:
            file = self.__collection / f"{doc['_id']}.json"
            file.unlink()

        return DeleteManyResult(deleted_count=len(docs))



import os
from pathlib import Path

import dotenv
import pytest
from redb.interface.errors import DocumentNotFound
from redb.interface.fields import Direction, SortColumn

from .utils import Embedding, RussianDog

dotenv.load_dotenv()


class TestmongoSystem:

    @pytest.fixture(scope="class", autouse=True)
    def clean_db(self):
        RussianDog.delete_many({})
        Embedding.delete_many({})
        yield
        RussianDog.delete_many({})
        Embedding.delete_many({})

    def test_insert_one(self):
        emb = Embedding(
            kb_name="KB",
            model="ai",
            text="Some data.",
            vector=[1, 2],
            source_url="www",
        )
        emb_id = Embedding.insert_one(emb)
        response = Embedding.find_one({"_id": emb_id.inserted_id})
        assert emb == response


    def test_find_one(self):
        vladmir = RussianDog(
            name="Vladimir",
            age=2,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=True,
        )
        vladmir_id = RussianDog.insert_one(vladmir)
        vladmir_from_db = RussianDog.find_one({"_id": vladmir_id.inserted_id})
        vladmir_breed = RussianDog.find_one(vladmir, fields=["breed"])
        assert vladmir_breed == {
            "breed": "Siberian Husky",
        }
        assert vladmir == vladmir_from_db

    def test_update_one(self):
        boris = RussianDog(
            name="Boris",
            age=2,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=True,
        )

        boris_id = RussianDog.insert_one(boris)
        RussianDog.update_one(filter=boris, update={"age": 3})
        boris.age = 3
        upsert_result = RussianDog.update_one(
            filter=boris.dict(exclude={"updated_at", "created_at"}),
            update={"is_good_boy": False},
            upsert=True,
        )
        boris_from_db = RussianDog.find_one({"_id": boris_id.inserted_id})

        assert upsert_result.matched_count == 1
        assert upsert_result.modified_count == 1
        assert boris_from_db.age == 3
        assert boris_from_db.is_good_boy == False

        irina = RussianDog(
            name="Irina",
            age=2,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=True,
        )
        upsert_result = RussianDog.update_one(
            filter=irina, update={"color": "White"}, upsert=True
        )
        irina.color = "White"
        irina_from_db = RussianDog.find_one({"_id": upsert_result.upserted_id})
        assert irina == irina_from_db
        assert upsert_result.matched_count == 0
        assert upsert_result.modified_count == 0

    def test_delete_one(self):
        svetlana = RussianDog(
            name="Svetlana",
            age=2,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=True,
        )
        svetlana_id = RussianDog.insert_one(svetlana)
        svetlana_from_db = RussianDog.find_one({"_id": svetlana_id.inserted_id})
        assert svetlana_from_db == svetlana

        RussianDog.delete_one(svetlana)
        with pytest.raises(DocumentNotFound):
            svetlana_from_db = RussianDog.find_one({"_id": svetlana_id.inserted_id})

    def test_insert_many(self):
        dog = RussianDog(
            name="dog", age=5, breed="Siberian Husky", color="Gray", is_good_boy=True
        )
        dogdog = RussianDog(
            name="dogdog",
            age=6,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=False,
        )
        ids = RussianDog.insert_many([dog, dogdog])
        print(ids.inserted_ids)
        dogs = RussianDog.find_many({"_id": {"$in": ids.inserted_ids}})
        dogs = sorted(dogs, key=lambda x: x.name)
        assert dogs == [dog, dogdog]
        RussianDog.delete_many({"_id": {"$in": ids.inserted_ids}})
        dict_ids = RussianDog.insert_many([dog.dict(), dogdog.dict()])
        dict_dogs = RussianDog.find_many({"_id": {"$in": dict_ids.inserted_ids}})
        dict_dogs = sorted(dict_dogs, key=lambda x: x.name)
        assert dict_dogs == [dog.dict(), dogdog.dict()]

    def test_find_many(self):
        natasha = RussianDog(
            name="Natasha",
            age=5,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=True,
        )
        oksana = RussianDog(
            name="Oksana",
            age=6,
            breed="Siberian Husky",
            color="Gray",
            is_good_boy=False,
        )
        ids = RussianDog.insert_many([natasha, oksana])
        dogs = RussianDog.find_many({"_id": {"$in": ids.inserted_ids}})
        dogs = sorted(dogs, key=lambda x: x.name)
        assert dogs == [natasha, oksana]
        # fields
        dog_ages = RussianDog.find_many(
            filter={"_id": {"$in": ids.inserted_ids}}, fields=["age"]
        )
        dog_ages = sorted(dog_ages, key=lambda x: x["age"])
        assert dog_ages[0]["age"] == 5 and dog_ages[1]["age"] == 6
        # sorted
        sorted_dogs = RussianDog.find_many(
            filter={"_id": {"$in": ids.inserted_ids}},
            sort=SortColumn(name="name", direction=Direction.ASCENDING),
        )
        assert sorted_dogs == [natasha, oksana]

    def test_update_many(self):
        oleg = RussianDog(
            name="Oleg",
            age=10,
            breed="Siberian Husky",
            color="Blue",
            is_good_boy=True,
        )
        katerina = RussianDog(
            name="Katerina",
            age=5,
            breed="Siberian Husky",
            color="Blue",
            is_good_boy=True,
        )
        # upadate
        ids = RussianDog.insert_many([oleg, katerina])
        update_result = RussianDog.update_many(
            filter={"color": "Blue"}, update={"color": "Red"}
        )
        katerina.color = "Red"
        oleg.color = "Red"
        dogs = RussianDog.find_many({"_id": {"$in": ids.inserted_ids}})
        dogs = sorted(dogs, key=lambda x: x.age)
        assert update_result.matched_count == 2
        assert update_result.modified_count == 2
        assert dogs == [katerina, oleg]

        # upsert
        romeo = RussianDog(
            name="Romeo",
            age=7,
            breed="Dachshund",
            color="Brown",
            is_good_boy=True,
        )
        upsert_result = RussianDog.update_many(
            filter=romeo,
            update={"color": "Black"},
            upsert=True,
        )
        romeo.color = "Black"

        upseted_dog = RussianDog.find_one(filter={"_id": upsert_result.upserted_id})
        assert upsert_result.matched_count == 0
        assert upsert_result.modified_count == 0
        assert upseted_dog == romeo

        # allow new fields
        update_allow_result = RussianDog.update_many(
            filter={"color": "Red"}, update={"is_pet": True}, allow_new_fields=True
        )
        allow_dogs = RussianDog.find_many({"_id": {"$in": ids.inserted_ids}})
        allow_dogs = sorted(allow_dogs, key=lambda x: x.age)
        katerina.is_pet = True
        oleg.is_pet = True

        assert update_allow_result.matched_count == 2
        assert update_allow_result.modified_count == 2
        assert allow_dogs[0] == katerina
        assert allow_dogs[1] == oleg

    def test_delete_many(self):
        viktor = RussianDog(
            name="Viktor",
            age=13,
            breed="Siberian Husky",
            color="White",
            is_good_boy=True,
        )
        alek = RussianDog(
            name="Alek",
            age=13,
            breed="Siberian Husky",
            color="White",
            is_good_boy=True,
        )
        ids = RussianDog.insert_many([viktor, alek])
        dogs = RussianDog.find_many({"_id": {"$in": ids.inserted_ids}})
        dogs = sorted(dogs, key=lambda x: x.name)
        assert [alek, viktor] == dogs

        RussianDog.delete_many(filter={"age": 13})
        with pytest.raises(DocumentNotFound):
            RussianDog.find_one({"_id": ids.inserted_ids[0]})
        with pytest.raises(DocumentNotFound):
            RussianDog.find_one({"_id": ids.inserted_ids[1]})



from dataclasses import dataclass, field
from typing import Any


@dataclass
class JSONConfig:
    client_folder_path: str
    default_database_folder_path: str | None = None


@dataclass
class MigoConfig:
    milvus_connection_alias: str
    milvus_host: str
    milvus_port: int
    mongo_database_uri: str
    mongo_kwargs: dict[str, Any] = field(default_factory=dict)
    milvus_kwargs: dict[str, Any] = field(default_factory=dict)


@dataclass
class MongoConfig:
    database_uri: str
    default_database: str | None = None
    driver_kwargs: dict = field(default_factory=dict)


CONFIGS = JSONConfig | MigoConfig | MongoConfig
CONFIG_TYPE = JSONConfig |  MigoConfig | MongoConfig | dict


def check_config(
    config: CONFIG_TYPE,
    base_class: CONFIGS,
) -> bool:
    return isinstance(config, base_class) or isinstance(config, dict)



from abc import ABC, abstractmethod
from typing import Sequence

from .database import Database


class Client(ABC):
    @abstractmethod
    def _get_driver_client(self) -> "Client":
        pass

    @abstractmethod
    def get_default_database(self) -> Database:
        pass

    @abstractmethod
    def get_databases(self) -> Sequence[Database]:
        pass

    @abstractmethod
    def get_database(self, name: str) -> Database:
        pass

    @abstractmethod
    def drop_database(self, name: str) -> bool:
        pass

    @abstractmethod
    def close(self) -> bool:
        pass



from dataclasses import dataclass
from enum import Enum
from types import UnionType
from typing import Any, ForwardRef, Literal, TypeVar, Union, _UnionGenericAlias

import pymongo
from bson import DBRef as BsonDBRef
from bson import ObjectId as BsonObjectId
from pydantic import BaseModel
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic.fields import ModelField
from pymongo.operations import (
    DeleteMany,
    DeleteOne,
    InsertOne,
    ReplaceOne,
    UpdateMany,
    UpdateOne,
)

T = TypeVar("T")

PyMongoOperations = TypeVar(
    "PyMongoOperations",
    bound=Union[
        InsertOne,
        DeleteOne,
        DeleteMany,
        ReplaceOne,
        UpdateOne,
        UpdateMany,
    ],
)


class Direction(Enum):
    ASCENDING = pymongo.ASCENDING
    DESCENDING = pymongo.DESCENDING
    GEO2D = pymongo.GEO2D
    GEOSPHERE = pymongo.GEOSPHERE
    HASHED = pymongo.HASHED
    TEXT = pymongo.TEXT

    @classmethod
    def from_string(
        cls, query: Literal["asc", "desc", "2d", "sphere", "hashed", "text"]
    ) -> "Direction":
        if query == "asc":
            return cls.ASCENDING
        elif query == "desc":
            return cls.DESCENDING
        elif query == "2d":
            return cls.GEO2D
        elif query == "sphere":
            return cls.GEOSPHERE
        elif query == "hashed":
            return cls.HASHED
        elif query == "text":
            return cls.TEXT
        else:
            raise ValueError(f"Invalid sort order: {query}.")


class Column(BaseModel):
    name: str


class IncludeColumn(Column):
    include: bool


class SortColumn(Column):
    direction: Direction


@dataclass
class Index:
    field: "ClassField"
    name: str | None = None
    unique: bool = False
    direction: Direction = Direction.ASCENDING
    extras: dict | None = None


@dataclass
class CompoundIndex:
    fields: list["ClassField"]
    name: str | None = None
    unique: bool = False
    direction: Direction = Direction.ASCENDING
    extras: dict | None = None


class FieldInfo(PydanticFieldInfo):
    def __init__(
        self,
        vector_type: str | None = None,
        dimensions: int | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.vector_type = vector_type
        self.dimensions = dimensions


def Field(*args, **kwargs) -> Any:
    return FieldInfo(*args, **kwargs)


class ClassField:
    def __init__(self, model_field: ModelField, base_class: BaseModel):
        self.model_field = model_field
        self.base_class = base_class
        self.attr_names = [model_field.alias]

    def resolve(self, obj: T) -> T | None:
        for attr_name in self.attr_names:
            if not obj:
                return obj
            if isinstance(obj, dict):
                get_attribute = lambda attr_name: obj[attr_name]
            else:
                get_attribute = lambda attr_name: getattr(obj, attr_name)

            if attr_name.endswith("[0]"):
                # If the attribute is iterable, we get its firts object
                obj = get_attribute(attr_name[: attr_name.rindex("[0]")])
                if not obj:
                    return obj
                obj = obj[0]
            else:
                # Otherwise just get it
                obj = get_attribute(attr_name)

        return obj

    def join_attrs(self, char: str = ".") -> str:
        return char.join(self.attr_names).replace("[0]", "")

    def __getattribute__(self, name: str) -> Any:
        if name in {
            "model_field",
            "base_class",
            "attr_names",
            "resolve",
            "join_attrs",
        }:
            return super().__getattribute__(name)

        try:
            annotation = _get_unwrapped_annotation(self.model_field.outer_type_)
            base_class = _get_type_from_annotation(annotation, self.base_class)

            if not hasattr(base_class, "__fields__"):
                raise AttributeError

            fields = base_class.__fields__
            if name not in fields:
                raise AttributeError

            self.model_field = fields[name]
            self.base_class = base_class
            self.attr_names.append(name)
            return self
        except AttributeError:
            return self.model_field.__getattribute__(name)

    def __getitem__(self, _):
        self.attr_names[-1] = f"{self.attr_names[-1]}[0]"
        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.attr_names})"


def _get_unwrapped_annotation(annotation: T) -> T:
    annotation = _unwrap_optional(annotation)
    annotation = _unwrap_union(annotation)
    annotation = _unwrap_iterable(annotation)
    annotation = _unwrap_union(annotation)
    return annotation


def _unwrap_optional(annotation: T) -> T:
    if hasattr(annotation, "_name") and annotation._name == "Optional":
        return annotation.__args__[0]
    return annotation


def _unwrap_iterable(annotation: T) -> T:
    if hasattr(annotation, "__args__"):
        if annotation.__origin__ not in {list, set}:
            raise ValueError("Are you trying to outsmart me!? HA HA")
        return annotation.__args__[0]
    return annotation


def _unwrap_union(annotation: T) -> T:
    if hasattr(annotation, "__args__") and annotation.__class__ in {
        UnionType,
        _UnionGenericAlias,
    }:
        return annotation.__args__[0]
    return annotation


def _get_type_from_annotation(annotation: T, base_class: BaseModel) -> BaseModel:
    if isinstance(annotation, ForwardRef):
        if not annotation.__forward_evaluated__:
            base_class.update_forward_refs()
        return annotation.__forward_value__
    return annotation


class ObjectId(BsonObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v) -> "ObjectId":
        if not BsonObjectId.is_valid(v):
            raise ValueError("Invalid ObjectId.")
        return ObjectId(v)

    @classmethod
    def __modify_schema__(cls, field_schema) -> str:
        field_schema.update(type="string")

    def __str__(self) -> str:
        return super().__str__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{str(self)}')"


class DBRefField(BaseModel):
    id: Any
    collection: str
    database: str | None = None

    class Config:
        json_encoders = {
            ObjectId: str,
        }

    def dict(self, *_, **__) -> dict:
        return {"$id": self.id, "$ref": self.collection, "$db": self.database}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.dict()}')"


class BsonDBRefField(BaseModel):
    id: Any = Field(alias="$id")
    ref: str = Field(alias="$ref")
    db: str | None = Field(None, alias="$db")


class DBRef(BsonDBRef):
    """Warning: does not work with openapi schema because of $ref."""

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        field = v
        if isinstance(v, BsonDBRef):
            field = DBRefField(
                id=v.id,
                collection=v.collection,
                database=v.database,
            )
        elif isinstance(field, dict):
            if "$id" in field:
                tmp = BsonDBRefField(**field)
                field = DBRefField(
                    id=tmp.id,
                    collection=tmp.ref,
                    database=tmp.db,
                )
            else:
                field = DBRefField(**v)

        if not isinstance(field, DBRefField):
            raise ValueError(f"Cannot construct PyDBField from value: {v}")

        return field

    @classmethod
    def __modify_schema__(cls, _):
        return DBRefField

    def __repr__(self):
        return f"{self.__class__.__name__}('{dict(self.as_doc())}')"



import shutil
from pathlib import Path

from redb.interface.database import Database

from .collection import JSONCollection


class JSONDatabase(Database):
    def __init__(self, database_path: Path) -> None:
        self.__database_folder_path = database_path

    def _get_driver_database(self) -> "JSONDatabase":
        return self

    def get_collections(self) -> list[JSONCollection]:
        return [
            JSONCollection(folder)
            for folder in self.__database_folder_path.glob("*")
            if folder.is_dir()
        ]

    def get_collection(self, name: str) -> JSONCollection:
        for folder in self.__database_folder_path.glob("*"):
            if folder.is_dir() and folder.name == name:
                return JSONCollection(folder)

        raise ValueError(f"Database {name} not found")

    def create_collection(self, name: str) -> None:
        (self.__database_folder_path / name).mkdir(exist_ok=True)

    def delete_collection(self, name: str) -> None:
        shutil.rmtree(self.get_collection(name).__class__.__name__)

    @property
    def name(self) -> str:
        return self.__database_folder_path.name

    def __getitem__(self, name: str) -> "JSONDatabase":
        return JSONCollection(self.__database_folder_path / name)

    def __truediv__(self, other: Path):
        return self.__database_folder_path / other



from typing import Sequence

from migo.client import Client as MigoDriverClient

from redb.interface.client import Client
from redb.interface.configs import MigoConfig

from .database import MigoDatabase


class MigoClient(Client):
    def __init__(self, migo_config: MigoConfig | dict):
        migo_config.mongo_kwargs["host"] = migo_config.mongo_database_uri
        migo_config.milvus_kwargs["alias"] = migo_config.milvus_connection_alias
        migo_config.milvus_kwargs["host"] = migo_config.milvus_host
        migo_config.milvus_kwargs["port"] = migo_config.milvus_port

        self.__migo_driver = MigoDriverClient(
            migo_config.mongo_kwargs, migo_config.milvus_kwargs
        )

    def _get_driver_client(self) -> MigoDriverClient:
        return self.__migo_driver

    def get_default_database(self) -> MigoDatabase:
        return MigoDatabase(self.__migo_driver.get_default_database())

    def get_database(self, name: str) -> MigoDatabase:
        return MigoDatabase(self.__migo_driver.get_database(name))

    def get_databases(self) -> Sequence[MigoDatabase]:
        return [
            MigoDatabase(database) for database in self.__migo_driver.get_databases()
        ]

    def get_database(self, name: str) -> MigoDatabase:
        return MigoDatabase(self.__migo_driver.get_database(name))

    def drop_database(self, name: str) -> bool:
        try:
            self.__migo_driver.drop_database(name)
            return True
        except:
            return False

    def close(self) -> bool:
        try:
            self.__migo_driver.close()
            return True
        except:
            return False



from dataclasses import dataclass
from typing import Any


@dataclass
class BulkWriteResult:
    deleted_count: int
    inserted_count: int
    matched_count: int
    modified_count: int
    upserted_count: int
    upserted_ids: int


@dataclass
class UpdateOneResult:
    matched_count: int
    modified_count: int
    upserted_id: Any


@dataclass
class UpdateManyResult(UpdateOneResult):
    pass


@dataclass
class ReplaceOneResult(UpdateOneResult):
    pass


@dataclass
class DeleteOneResult:
    deleted_count: int


@dataclass
class DeleteManyResult:
    deleted_count: int


@dataclass
class InsertManyResult:
    inserted_ids: list[Any]


@dataclass
class InsertOneResult:
    inserted_id: Any



from pymongo.database import Database as PymongoDatabase
from pymongo.errors import CollectionInvalid

from redb.interface.database import Database

from .collection import MongoCollection


class MongoDatabase(Database):
    def __init__(self, database: PymongoDatabase) -> None:
        self.__database = database

    def _get_driver_database(self) -> PymongoDatabase:
        return self.__database

    def get_collections(self) -> list[MongoCollection]:
        return [
            MongoCollection(self.__database[col["name"]])
            for col in self.__database.list_collections()
        ]

    def get_collection(self, name: str) -> MongoCollection:
        return MongoCollection(self.__database[name])

    def create_collection(self, name: str) -> bool:
        try:
            self.__database.create_collection(name)
            return True
        except CollectionInvalid:
            return False

    def delete_collection(self, name: str) -> bool:
        try:
            self.__database.drop_collection(name)
            return True
        except CollectionInvalid:
            return False
        
    @property
    def name(self) -> str:
        return self.__database.name

    def __getitem__(self, name: str) -> MongoCollection:
        return self.get_collection(name)



import shutil
from pathlib import Path

from redb.interface.client import Client
from redb.interface.configs import JSONConfig

from .database import JSONDatabase


class JSONClient(Client):
    def __init__(self, json_config: JSONConfig | dict) -> None:
        if isinstance(json_config, dict):
            json_config = JSONConfig(**json_config)
        self.__client_folder_path = Path(json_config.client_folder_path)

        database_default_path = json_config.default_database_folder_path
        if database_default_path is None:
            self.__default_database = JSONDatabase(
                self.__client_folder_path / Path("default-database")
            )
        else:
            self.__default_database = JSONDatabase(
                self.__client_folder_path / database_default_path
            )

    def _get_driver_client(self) -> "JSONClient":
        return self

    def get_databases(self) -> list[JSONDatabase]:
        return [
            JSONDatabase(folder)
            for folder in self.__client_folder_path.glob("*")
            if folder.is_dir()
        ]

    def get_database(self, name: str) -> JSONDatabase:
        for folder in self.__client_folder_path.glob("*"):
            if folder.is_dir() and folder.name == name:
                return JSONDatabase(folder)

        raise ValueError(f"Database {name} not found")

    def get_default_database(self) -> JSONDatabase:
        return self.__default_database

    def drop_database(self, name: str) -> bool:
        try:
            shutil.rmtree(self.get_database(name).database_path)
            return True
        except:
            return False

    def close(self) -> bool:
        return True

    def __truediv__(self, other):
        return self.__client_folder_path / other



from typing import Sequence

from pymongo import MongoClient as PymongoClient

from redb.interface.client import Client
from redb.interface.configs import MongoConfig

from .database import MongoDatabase


class MongoClient(Client):
    def __init__(self, mongo_config: MongoConfig | dict):
        if isinstance(mongo_config, dict):
            mongo_config = MongoConfig(**mongo_config)
        self.__client = PymongoClient(
            mongo_config.database_uri, **mongo_config.driver_kwargs
        )
        if mongo_config.default_database is None:
            self.__default_database = MongoDatabase(
                self.__client.get_default_database()
            )
        else:
            self.__default_database = self.get_database(mongo_config.default_database)

    def _get_driver_client(self) -> PymongoClient:
        return self.__client

    def get_databases(self) -> Sequence[MongoDatabase]:
        return [
            MongoDatabase(self.__client.get_database(database["name"]))
            for database in self.__client.list_databases()
        ]

    def get_database(self, name: str) -> MongoDatabase:
        return MongoDatabase(self.__client.get_database(name))

    def get_default_database(self) -> MongoDatabase:
        return self.__default_database

    def drop_database(self, name: str) -> bool:
        try:
            self.__client.drop_database(name)
            return True
        except TypeError:
            return False

    def close(self) -> bool:
        try:
            self.__client.close()
            return True
        except Exception:
            return False



from abc import ABC, abstractmethod

from .collection import Collection


class Database(ABC):
    @abstractmethod
    def _get_driver_database(self) -> "Database":
        pass

    @abstractmethod
    def get_collections(self) -> list[Collection]:
        pass

    @abstractmethod
    def get_collection(self, name: str) -> Collection:
        pass

    @abstractmethod
    def create_collection(self, name: str) -> bool:
        pass

    @abstractmethod
    def delete_collection(self, name: str) -> bool:
        pass

    @abstractmethod
    def __getitem__(self, name: str) -> Collection:
        pass



import os
from operator import attrgetter

import pytest
from redb.behaviors import SoftDeletinDoc
from redb.interface.errors import DocumentNotFound
from redb.interface.fields import ClassField
from redb.core import RedB
from redb.interface.configs import MongoConfig

class Cat(SoftDeletinDoc):
    name: str

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        return [cls.name]  # type: ignore


@pytest.fixture(scope="module", autouse=True)
def client():
    RedB.setup(
        MongoConfig(
            database_uri=os.environ["MONGODB_URI"],
        )
    )


def test_soft_deletion():
    obj = Cat(name="Fluffy")
    filters = dict(_id=obj.id)
    try:
        obj.insert()
        Cat.soft_delete_one(filters)
        with pytest.raises(DocumentNotFound):
            Cat.find_one(filters)
    finally:
        Cat.delete_one(filters)


def test_soft_undeletion():
    obj = Cat(name="Fluffy")
    filters = dict(_id=obj.id)
    try:
        obj.insert()
        Cat.soft_delete_one(filters)
        with pytest.raises(DocumentNotFound):
            Cat.find_one(filters)
            
        Cat.soft_undelete_one(filters)
        cat = Cat.find_one(filters)
        assert cat.dict(exclude={"updated_at"}) == obj.dict(exclude={"updated_at"})
    finally:
        Cat.delete_one(filters)


def test_soft_delete_then_undelete_many():
    objs = [Cat(name="Fluffy"), Cat(name="Whiskers")]
    filters = {"_id": {"$in": [o.id for o in objs]}}
    try:
        Cat.insert_many(objs)
        Cat.soft_delete_many(filters)
        found_objs = Cat.find_many(filters)
        assert len(found_objs) == 0
        
        Cat.soft_undelete_many(filters)
        found_objs = Cat.find_many(filters)
        found_objs = sorted(found_objs, key=attrgetter("name"))
        assert len(objs) == len(found_objs)
        for expec, found in zip(objs, found_objs):
            assert expec.dict(exclude={"updated_at"}) == found.dict(exclude={"updated_at"})

    finally:
        Cat.delete_many(filters)



import pydantic
from pydantic import BaseModel
from bson import ObjectId as BsonObjectId, DBRef as BsonDBRef


class ObjectId(BsonObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, v):
        if not ObjectId.is_valid(v):
            raise ValueError("Invalid ObjectId.")
        return ObjectId(v)
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type="string")

    def __repr__(self):
        return f"{self.__class__.__name__}('{str(self)}')"


class DBRefField(BaseModel):
    collection:str
    id: ObjectId
    database: str | None = None

    class Config:
        json_encoders = {
            ObjectId: str,
        }


class DBRef(BsonDBRef):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        field = v
        if isinstance(field, dict):
            field = DBRefField(**v)

        if not isinstance(field, DBRefField):
            raise ValueError(f"Cannot construct PyDBField from value: {v}")

        return field

    @classmethod
    def __modify_schema__(cls, _):
        return DBRefField

    def __repr__(self):
        return f"{self.__class__.__name__}('{dict(self.as_doc())}')"




class DemoModel(pydantic.BaseModel):
    a: str
    b: ObjectId
    c: DBRef


    class Config:
        json_encoders = {
            DBRef: lambda d: dict(d.as_doc()),
            ObjectId: str,
        }


a = DemoModel(
    a="",
    b=ObjectId(),
    c={"collection": "batata", "id": ObjectId()},
)
print(a)
print(a.dict())
print(a.json())

print(DemoModel(
    a="",
    b=str(ObjectId()),
    c=DBRefField(collection="batata", id=str(ObjectId()))
))



from redb.core.document import Document, _get_return_cls, _format_fields
from redb.interface.fields import IncludeColumn


class Scarecrow(Document):
    name: str
    other: None = None


# _format_fields([IncludeColumn(name="name", include=True), IncludeColumn(name="other", include=False)])
def test_return_cls():
    klass = _get_return_cls(Scarecrow, {"name": True, "other": False})
    assert klass is Scarecrow
    klass = _get_return_cls(Scarecrow, {"name": False, "other": False})
    assert klass is dict  # name is unselected
    klass = _get_return_cls(Scarecrow, {"name": False, "other": True})
    assert klass is dict  # name is unselected
    klass = _get_return_cls(Scarecrow, {"name": True, "other": True})
    assert klass is Scarecrow
    klass = _get_return_cls(Scarecrow, {"name": True})
    assert klass is Scarecrow
    klass = _get_return_cls(Scarecrow, {"other": False})
    assert klass is Scarecrow
    klass = _get_return_cls(Scarecrow, {"other": True})
    assert klass is dict  # missing name required
    klass = _get_return_cls(Scarecrow, {"name": False})
    assert klass is dict  # name is unselected



from migo.database import Database as MigoDriverDatabase

from redb.interface.database import Database

from .collection import MigoCollection


class MigoDatabase(Database):
    def __init__(self, database: MigoDriverDatabase) -> None:
        self.__database = database

    def _get_driver_database(self) -> MigoDriverDatabase:
        return self.__database

    def get_collections(self) -> list[MigoCollection]:
        return [
            MigoCollection(collection)
            for collection in self.__database.get_collections()
        ]

    def get_collection(self, name: str) -> MigoCollection:
        return MigoCollection(self.__database.get_collection(name))

    def create_collection(self, name: str) -> bool:
        try:
            self.__database.create_collection(name)
            return True
        except:
            return False

    def delete_collection(self, name: str) -> bool:
        try:
            self.__database.delete_collection(name)
            return True
        except:
            return False
        
    @property
    def name(self) -> str:
        return self.__database.name

    def __getitem__(self, name: str) -> MigoCollection:
        return self.get_collection(name)



from datetime import datetime
from typing import Any, Dict, Sequence, Type, TypeAlias, TypeVar, Union, cast

import pytz
from pymongo.errors import DuplicateKeyError
from redb.interface.errors import (
    CannotUpdateIdentifyingField,
    UniqueConstraintViolation,
)
from redb.interface.fields import (
    CompoundIndex,
    DBRef,
    Field,
    IncludeColumn,
    Index,
    ObjectId,
    PyMongoOperations,
    SortColumn,
)
from redb.interface.results import (
    BulkWriteResult,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)

from .base import BaseDocument

DocumentData: TypeAlias = Union["Document", Dict[str, Any]]
IncludeColumns: TypeAlias = list[IncludeColumn] | list[str] | None
OptionalDocumentData: TypeAlias = Union["Document", dict[str, Any], None]
SortColumns: TypeAlias = list[SortColumn] | SortColumn | None
T = TypeVar("T", bound="Document")


class Document(BaseDocument):
    id: str = Field(alias="_id")  # type: ignore
    created_at: datetime = Field(default_factory=lambda: datetime.now(pytz.UTC))  # type: ignore
    updated_at: datetime = Field(default_factory=lambda: datetime.now(pytz.UTC))  # type: ignore

    class Config:
        json_encoders = {
            datetime: lambda d: d.isoformat(),
            DBRef: lambda ref: dict(ref.as_doc()),
            ObjectId: str,
        }
        smart_union = True

    def __init__(self, **data: Any) -> None:
        calculate_hash = False
        if "id" in data:
            data["_id"] = data.pop("id")
        if "_id" not in data:
            data["_id"] = None
            calculate_hash = True

        data = self.update_kwargs(data)
        if calculate_hash:
            data["_id"] = self.get_hash(data)
        super().__init__(**data)

    @classmethod
    def create_indexes(cls: Type[T]) -> None:
        collection = Document._get_collection(cls)
        indexes = cls.get_indexes()
        for index in indexes:
            index = _format_index(index)
            collection.create_index(index)

    def find(
        self: T,
        fields: IncludeColumns = None,
        skip: int = 0,
    ) -> T:
        collection = Document._get_collection(self.__class__)
        filter = _format_document_data(self)
        formatted_fields = _format_fields(fields)
        return_cls = _get_return_cls(self.__class__, fields)
        return collection.find_one(
            cls=self.__class__,
            return_cls=return_cls,
            filter=filter,
            skip=skip,
            fields=formatted_fields,
        )

    @classmethod
    def find_one(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        skip: int = 0,
    ) -> T:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        formatted_fields = _format_fields(fields)
        return_cls = _get_return_cls(cls, formatted_fields)
        return collection.find_one(
            cls=cls,
            return_cls=return_cls,
            filter=filter,
            skip=skip,
            fields=formatted_fields,
        )

    @classmethod
    def find_many(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        sort: SortColumns = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[T]:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        formatted_fields = _format_fields(fields)
        return_cls = _get_return_cls(cls, formatted_fields)
        sort_order = _format_sort(sort)
        return collection.find(
            cls=cls,
            return_cls=return_cls,
            filter=filter,
            fields=formatted_fields,
            sort=sort_order,
            skip=skip,
            limit=limit,
        )

    @classmethod
    def distinct(
        cls: Type[T],
        key: str,
        filter: OptionalDocumentData = None,
    ) -> list[Any]:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        return collection.distinct(
            cls=cls,
            key=key,
            filter=filter,
        )

    @classmethod
    def count_documents(
        cls: Type[T],
        filter: OptionalDocumentData = None,
    ) -> int:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        return collection.count_documents(
            cls=cls,
            filter=filter,
        )

    @classmethod
    def bulk_write(
        cls: Type[T],
        operations: list[PyMongoOperations],
    ) -> BulkWriteResult:
        collection = Document._get_collection(cls)
        return collection.bulk_write(
            cls=cls,
            operations=operations,
        )

    def insert(self: T) -> InsertOneResult:
        collection = Document._get_collection(self.__class__)
        data = _format_document_data(self)
        try:
            return collection.insert_one(
                cls=self.__class__,
                data=data,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=self.collection_name()
            )

    @classmethod
    def insert_one(
        cls: Type[T],
        data: DocumentData,
    ) -> InsertOneResult:
        _validate_fields(cls, data)

        collection = Document._get_collection(cls)
        data = _format_document_data(data)
        try:
            return collection.insert_one(
                cls=cls,
                data=data,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=cls.collection_name()
            )

    @classmethod
    def insert_vectors(
        cls: Type[T],
        data: Dict[str, list[Any]],
    ) -> InsertManyResult:
        collection = Document._get_collection(cls)
        keys = list(data.keys())
        values_size = len(data[keys[0]])
        instances = [None] * values_size
        instances = [{key: data[key][i] for key in keys} for i in range(values_size)]
        try:
            return collection.insert_many(
                cls=cls,
                data=instances,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=cls.collection_name()
            )

    @classmethod
    def insert_many(
        cls: Type[T],
        data: Sequence[DocumentData],
    ) -> InsertManyResult:
        for val in data:
            _validate_fields(cls, val)
        collection = Document._get_collection(cls)
        data = [_format_document_data(val) for val in data]
        try:
            result = collection.insert_many(
                cls=cls,
                data=data,
            )
            return result
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=cls.collection_name()
            )

    def replace(
        self: T,
        replacement: DocumentData,
        upsert: bool = False,
        allow_new_fields: bool = False,
    ) -> ReplaceOneResult:
        if not allow_new_fields:
            _validate_fields(self.__class__, replacement)

        collection = Document._get_collection(self.__class__)
        filter = _format_document_data(self)
        replacement = _format_document_data(replacement)
        return collection.replace_one(
            cls=self.__class__,
            filter=filter,
            replacement=replacement,
            upsert=upsert,
        )

    @classmethod
    def replace_one(
        cls: Type[T],
        filter: DocumentData,
        replacement: DocumentData,
        upsert: bool = False,
        allow_new_fields: bool = False,
    ) -> ReplaceOneResult:
        if not allow_new_fields:
            _validate_fields(filter.__class__, replacement)

        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        replacement = _format_document_data(replacement)
        return collection.replace_one(
            cls=cls,
            filter=filter,
            replacement=replacement,
            upsert=upsert,
        )

    def update(
        self,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
    ) -> UpdateOneResult:
        if not allow_new_fields:
            _validate_fields(self.__class__, update)

        collection = Document._get_collection(self.__class__)
        filter = _format_document_data(self)
        update = _format_document_data(update)

        if not upsert:
            filter = _optimize_filter(self.__class__, filter)

        _raise_if_updating_hashable(self.__class__, update)
        if operator is not None:
            update = {operator: update}

        result = collection.update_one(
            cls=self.__class__,
            filter=filter,
            update=update,
            upsert=upsert,
        )
        collection.update_one(
            cls=self.__class__,
            filter=filter,
            update={"$set": {"updated_at": datetime.now(pytz.UTC).isoformat()}},
        )
        return result

    @classmethod
    def update_one(
        cls,
        filter: DocumentData,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
    ) -> UpdateOneResult:
        if not allow_new_fields:
            _validate_fields(cls, update)

        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        update_data = _format_document_data(update)

        if not upsert:
            filter = _optimize_filter(cls, filter)

        _raise_if_updating_hashable(cls, update_data)
        if operator is not None:
            update_data = {operator: update_data}

        try:
            result = collection.update_one(
                cls=cls,
                filter=filter,
                update=update_data,
                upsert=upsert,
            )
            collection.update_one(
                cls=cls,
                filter=filter,
                update={"$set": {"updated_at": datetime.now(pytz.UTC).isoformat()}},
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=cls.collection_name()
            )
        return result

    @classmethod
    def update_many(
        cls: Type[T],
        filter: DocumentData,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
    ) -> UpdateManyResult:
        if not allow_new_fields:
            _validate_fields(cls, update)

        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        update = _format_document_data(update)

        if not upsert:
            filter = _optimize_filter(cls, filter)

        _raise_if_updating_hashable(cls, update)
        if operator is not None:
            update = {operator: update}

        try:
            result = collection.update_many(
                cls=cls,
                filter=filter,
                update=update,
                upsert=upsert,
            )
            collection.update_many(
                cls=cls,
                filter=filter,
                update={"$set": {"updated_at": datetime.now(pytz.UTC).isoformat()}},
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"], collection_name=cls.collection_name()
            )

        return result

    def delete(self: T) -> DeleteOneResult:
        collection = Document._get_collection(self.__class__)
        filter = _format_document_data(self)
        return collection.delete_one(
            cls=self.__class__,
            filter=filter,
        )

    @classmethod
    def delete_one(
        cls: Type[T],
        filter: DocumentData,
    ) -> DeleteOneResult:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        return collection.delete_one(
            cls=cls,
            filter=filter,
        )

    @classmethod
    def delete_many(
        cls: Type[T],
        filter: DocumentData,
    ) -> DeleteManyResult:
        collection = Document._get_collection(cls)
        filter = _format_document_data(filter)
        return collection.delete_many(
            cls=cls,
            filter=filter,
        )


def _validate_fields(cls: Type[DocumentData], data: DocumentData) -> None:
    if data is None:
        return

    if not isinstance(cls, BaseDocument):
        return

    class_fields = cls.__fields__
    for key in data:
        if key not in class_fields:
            raise ValueError(f"Key {key} is not present in the original document")


def _get_return_cls(
    cls: Type[Document],
    fields: dict[str, bool] | None = None,
) -> Type[Document | dict]:
    if not fields:
        return cls
    return_type = cls
    selected_fields = {k for k, v in fields.items() if v}
    unselected_fields = {k for k, v in fields.items() if not v}
    for v in cls.__fields__.values():
        if not v.required:
            continue
        if v.alias == "_id":
            continue
        if unselected_fields and v.alias in unselected_fields:
            return_type = dict
            break
        if selected_fields and v.alias not in selected_fields:
            return_type = dict
            break
    return return_type


def _format_fields(fields: IncludeColumns) -> dict[str, bool] | None:
    if fields is None:
        return None
    if all(isinstance(f, str) for f in fields):
        fields = cast(list[str], fields)
        formatted_fields = {field: True for field in fields}
    else:
        fields = cast(list[IncludeColumn], fields)
        formatted_fields = {field.name: field.include for field in fields}
    if any(v for v in formatted_fields.values()):
        # if you're choosing which fields you want, we'll pass _id = False
        # otherwise (if you're only choosing which ones you don't want),
        # we'll let it be included by default.
        if "_id" not in formatted_fields:
            formatted_fields["_id"] = False
    return formatted_fields


def _format_sort(sort: SortColumns) -> list[tuple[str, str | int]] | None:
    if sort is None:
        return sort
    if isinstance(sort, list):
        formatted_sort = [(field.name, field.direction.value) for field in sort]
    else:
        formatted_sort = [(sort.name, sort.direction.value)]
    return formatted_sort


def _format_document_data(data: OptionalDocumentData) -> dict[str, Any]:
    if data is None:
        return {}
    if isinstance(data, BaseDocument):
        return data.dict(by_alias=True)
    return data


def _format_index(index: Index | CompoundIndex):
    if isinstance(index, Index):
        index = CompoundIndex(
            fields=[index.field],
            name=index.name,
            unique=index.unique,
            direction=index.direction,
        )

    return index


def _raise_if_updating_hashable(cls: Type[T], update_dict: dict):
    hashable_field_attr_names = set(
        x.model_field.name for x in cls.get_hashable_fields()
    )
    for field in update_dict.keys():
        if field in hashable_field_attr_names:
            m = f"Cannot update hashable field {field!r} on {cls.__name__!r}."
            raise CannotUpdateIdentifyingField(m)


def _optimize_filter(cls: Type[T], filter: dict) -> dict:
    if "_id" in filter:
        return {"_id": filter["_id"]}
    unique_fields = set()
    indexes = [i for i in cls.get_indexes() if isinstance(i, Index)]
    for index in indexes:
        if index.unique:
            unique_fields.add(index.field.model_field.alias)
    for key, value in filter.items():
        if key in unique_fields:
            return {key: value}
    compound_indexes = [i for i in cls.get_indexes() if isinstance(i, CompoundIndex)]
    for c_index in compound_indexes:
        if c_index.unique:
            if all(k in filter for k in c_index.fields):
                return {k: filter[k] for k in c_index.fields}
    return filter



import types
from enum import Enum
from typing import Any, Type, TypeVar

from pydantic import BaseModel
from pydantic.fields import ModelField

T = TypeVar("T", bound=BaseModel)

DEFAULT_CONTAINER_VALUES = {"list": [], "tuple": (), "set": set()}
DEFAULT_PRIMITIVE_VALUES = {
    "str": "string",
    "float": 0.0,
    "int": 0,
    "bool": False,
    "None": None,
    "bytes": b"",
}
DEFAULT_VALUES = DEFAULT_PRIMITIVE_VALUES | DEFAULT_CONTAINER_VALUES


def generate_examples(pydantic_models: list[Type[T]]) -> dict[str, dict[str, Any]]:
    out = {}
    for model in pydantic_models:
        example = get_example_from_pydantic(model)
        out[example["summary"]] = example

    return out


def get_example_from_pydantic(pydantic_model: Type[T]) -> dict[str, Any]:
    class_name = pydantic_model.__name__
    class_doc = pydantic_model.__doc__

    summary = _get_summary(class_name)
    description = class_doc.strip() if class_doc is not None else ""
    value = _get_value(pydantic_model)

    return {"summary": summary, "description": description, "value": value}


def _get_summary(class_name: str) -> str:
    summary = class_name[0]
    for char in class_name[1:]:
        if char.isupper():
            summary += " "
        summary += char

    return summary


def _get_value(pydantic_model: Type[T]) -> dict:
    value = {}
    for name, field in pydantic_model.__fields__.items():
        if field.default is not None or field.default_factory is not None:
            field_value = field.default
            if field_value is None and field.default_factory is not None:
                field_value = field.default_factory()

            if isinstance(field_value, Enum):
                field_value = field_value.value
            elif isinstance(field_value, BaseModel):
                field_value = _get_value(field_value.__class__)

            value[name] = field_value
            continue

        value[name] = _parse_value(field)

    return value


def _parse_value(field: ModelField):
    value = None
    if isinstance(field.type_, types.UnionType):
        field.type_ = field.type_.__args__[0]

    if isinstance(field.outer_type_, types.UnionType):
        field.outer_type_ = field.outer_type_.__args__[0]

    if field.type_.__name__ in DEFAULT_CONTAINER_VALUES:
        try:
            tmp = field.type_
            field.type_ = field.type_.__args__[0]
            field.outer_type_ = tmp
        except:
            return DEFAULT_CONTAINER_VALUES[field.type_.__name__]

    if field.type_.__name__ in DEFAULT_PRIMITIVE_VALUES:
        value = DEFAULT_PRIMITIVE_VALUES[field.type_.__name__]
    elif issubclass(field.type_, BaseModel):
        value = _get_value(field.type_)
    elif issubclass(field.type_, Enum):
        value = next(iter(field.type_)).value
    else:
        value = field.type_.__name__

    if field.outer_type_.__name__ in DEFAULT_CONTAINER_VALUES:
        value = _build_container(value, field)
    
    return value
        


def _build_container(value: Any, field: ModelField):
    if field.outer_type_.__name__ == "list":
        return [value] if value is not None else []
    elif field.outer_type_.__name__ == "set":
        return {value,} if value is not None else set()
    elif field.outer_type_.__name__ == "tuple":
        return (value,) if value is not None else tuple()
    return value


from .document import BaseDocument, Document
from .instance import RedB
from redb.interface.fields import CompoundIndex, Field, Index, ClassField
from redb.interface.configs import JSONConfig, MongoConfig, MigoConfig



import contextlib
from datetime import datetime
from typing import Any, ContextManager, Dict, Sequence, Type, TypeVar, overload

import pytz
from pymongo.errors import DuplicateKeyError

from redb.core.document import (
    Document,
    DocumentData,
    IncludeColumns,
    OptionalDocumentData,
    SortColumns,
    _format_document_data,
    _format_fields,
    _format_index,
    _format_sort,
    _get_return_cls,
    _optimize_filter,
    _raise_if_updating_hashable,
    _validate_fields,
)
from redb.interface.collection import (
    BulkWriteResult,
    Collection,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    PyMongoOperations,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)
from redb.interface.configs import (
    CONFIG_TYPE,
    JSONConfig,
    MigoConfig,
    MongoConfig,
    check_config,
)
from redb.interface.errors import UniqueConstraintViolation

T = TypeVar("T", bound=Document)


class CollectionWrapper:
    def __init__(self, collection: Collection, collection_class: T) -> None:
        self.__collection = collection
        self.__collection_class = collection_class

    def _get_driver_collection(self) -> Any:
        return self.__collection._get_driver_collection()

    def create_indexes(self) -> None:
        indexes = self.__collection_class.get_indexes()
        for index in indexes:
            index = _format_index(index)
            self.__collection.create_index(index)

    def find_one(
        self,
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        skip: int = 0,
    ) -> "Document":
        chosen_fields = _format_fields(fields)
        return_cls = _get_return_cls(self.__collection_class, chosen_fields)
        filters = _format_document_data(filter)
        return self.__collection.find_one(
            cls=self.__collection_class,
            return_cls=return_cls,
            filter=filters,
            skip=skip,
            fields=chosen_fields,
        )

    def find_many(
        self,
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        sort: SortColumns = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list["Document"]:
        chosen_fields = _format_fields(fields)
        return_cls = _get_return_cls(self.__collection_class, chosen_fields)
        filter = _format_document_data(filter)
        sort = _format_sort(sort)
        return self.__collection.find(
            cls=self.__collection_class,
            return_cls=return_cls,
            filter=filter,
            fields=chosen_fields,
            sort=sort,
            skip=skip,
            limit=limit,
        )

    def distinct(
        self,
        key: str,
        filter: OptionalDocumentData = None,
    ) -> list["Document"]:
        filter = _format_document_data(filter)
        return self.__collection.distinct(
            cls=self.__collection_class,
            key=key,
            filter=filter,
        )

    def count_documents(
        self,
        filter: OptionalDocumentData = None,
    ) -> int:
        filter = _format_document_data(filter)
        return self.__collection.count_documents(
            cls=self.__collection_class,
            filter=filter,
        )

    def bulk_write(
        self,
        operations: list[PyMongoOperations],
    ) -> BulkWriteResult:
        return self.__collection.bulk_write(
            cls=self.__collection_class,
            operations=operations,
        )

    def insert_one(self, data: DocumentData) -> InsertOneResult:
        data = _format_document_data(data)

        try:
            return self.__collection.insert_one(
                cls=self.__collection_class,
                data=data,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(dup_keys=e.details["keyValue"])

    def insert_vectors(self, data: Dict[str, list[Any]]) -> InsertManyResult:
        keys = list(data.keys())
        values_size = len(data[keys[0]])
        instances = [None] * values_size
        instances = [{key: data[key][i] for key in keys} for i in range(values_size)]

        try:
            return self.__collection.insert_many(
                cls=self.__collection_class,
                data=instances,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(dup_keys=e.details["keyValue"])

    def insert_many(
        self,
        data: Sequence[DocumentData],
    ) -> InsertManyResult:
        [_validate_fields(self.__collection_class, val) for val in data]

        data = [_format_document_data(val) for val in data]

        try:
            return self.__collection.insert_many(
                cls=self.__collection_class,
                data=data,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(dup_keys=e.details["keyValue"])

    def replace_one(
        self,
        filter: DocumentData,
        replacement: DocumentData,
        upsert: bool = False,
        allow_new_fields: bool = False,
    ) -> ReplaceOneResult:
        if not allow_new_fields:
            _validate_fields(self.__collection_class, replacement)

        filter = _format_document_data(filter)
        replacement = _format_document_data(replacement)
        return self.__collection.replace_one(
            cls=self.__collection_class,
            filter=filter,
            replacement=replacement,
            upsert=upsert,
        )

    def update_one(
        self,
        filter: DocumentData,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
    ) -> UpdateOneResult:
        if not allow_new_fields:
            _validate_fields(self.__collection_class, update)

        filter = _format_document_data(filter)
        update = _format_document_data(update)

        if not upsert:
            filter = _optimize_filter(self.__collection_class, filter)

        _raise_if_updating_hashable(self.__collection_class, update)
        if operator is not None:
            update = {operator: update}

        result = self.__collection.update_one(
            cls=self.__collection_class,
            filter=filter,
            update=update,
            upsert=upsert,
        )
        self.__collection.update_one(
            cls=self.__collection_class,
            filter=filter,
            update={"$set": {"updated_at": datetime.now(pytz.UTC).isoformat()}},
        )
        return result

    def update_many(
        self,
        filter: DocumentData,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
    ) -> UpdateManyResult:
        if not allow_new_fields:
            _validate_fields(self.__collection_class, update)

        filter = _format_document_data(filter)
        update = _format_document_data(update)

        if not upsert:
            filter = _optimize_filter(self.__collection_class, filter)

        _raise_if_updating_hashable(self.__collection_class, update)
        if operator is not None:
            update = {operator: update}

        result = self.__collection.update_many(
            cls=self.__collection_class,
            filter=filter,
            update=update,
            upsert=upsert,
        )
        self.__collection.update_many(
            cls=self.__collection_class,
            filter=filter,
            update={"$set": {"updated_at": datetime.now(pytz.UTC).isoformat()}},
        )
        return result

    def delete_one(self, filter: DocumentData) -> DeleteOneResult:
        filter = _format_document_data(filter)
        return self.__collection.delete_one(
            cls=self.__collection_class,
            filter=filter,
        )

    def delete_many(self, filter: DocumentData) -> DeleteManyResult:
        filter = _format_document_data(filter)
        return self.__collection.delete_many(
            cls=self.__collection_class,
            filter=filter,
        )


@overload
def transaction(
    collection: Type[Document],
    backend: str | None = None,
    config: CONFIG_TYPE | None = None,
    db_name: str | None = None,
) -> ContextManager[CollectionWrapper]:
    pass


@overload
def transaction(
    collection: str,
    backend: str | None = None,
    config: CONFIG_TYPE | None = None,
    db_name: str | None = None,
) -> ContextManager[Collection]:
    pass


@contextlib.contextmanager
def transaction(
    collection: Document | str,
    backend: str | None = None,
    config: CONFIG_TYPE | None = None,
    db_name: str | None = None,
) -> Collection | CollectionWrapper:
    new_client, client = get_client(backend, config)

    if db_name is None:
        database = client.get_default_database()
    else:
        database = client.get_database(db_name)

    if isinstance(collection, str):
        collection = database.get_collection(collection)
    else:
        collection_name = collection.collection_name()
        driver_collection = database.get_collection(collection_name)
        collection = CollectionWrapper(driver_collection, collection)

    yield collection

    if new_client:
        client.close()


def get_client(backend: str | None, config: CONFIG_TYPE):
    if backend == "json" or (backend is None and check_config(config, JSONConfig)):
        from redb.json_system import JSONClient

        return True, JSONClient(config)

    elif backend == "mongo" or (backend is None and check_config(config, MongoConfig)):
        from redb.mongo_system import MongoClient

        return True, MongoClient(config)

    elif backend == "migo" and (backend is None and check_config(config, MigoConfig)):
        from redb.migo_system import MigoClient

        return True, MigoClient(config)

    else:
        from redb.core.instance import RedB

        return False, RedB.get_client()



import hashlib
import dataclasses
from typing import Any, Callable, ClassVar, Dict, Type

from pydantic import BaseModel, Field
from pydantic.main import ModelMetaclass
from typing_extensions import dataclass_transform

from redb.interface.fields import ClassField, CompoundIndex, Index

from .instance import RedB

IMPORT_ERROR_MSG = (
    "%s does not seem to be installed, maybe you forgot to `pip install redb[%s]`"
)

@dataclass_transform(kw_only_default=True, field_specifiers=(Field, ClassField))
class DocumentMetaclass(ModelMetaclass):
    def __getattribute__(cls_or_self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            if hasattr(cls_or_self, "__fields__"):
                if name not in cls_or_self.__fields__:
                    raise e

                return ClassField(
                    model_field=cls_or_self.__fields__[name],
                    base_class=cls_or_self,
                )


class BaseDocument(BaseModel, metaclass=DocumentMetaclass):
    __database_name__: ClassVar[str | None] = None

    def dict(self, *args, **kwargs) -> dict:
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        out = super().dict(*args, **kwargs)
        return _apply_encoders(out, self.__config__.json_encoders)

    @staticmethod
    def _get_driver_collection(
        instance_or_class: Type["BaseDocument"] | "BaseDocument",
    ) -> Any:
        collection_name = instance_or_class.collection_name()
        database_name = instance_or_class.__database_name__

        client = RedB.get_client()
        database = (
            client.get_database(database_name)
            if database_name
            else client.get_default_database()
        )

        client_name = RedB.get_client_name()
        if client_name == "mongo":
            return database._get_driver_database()[collection_name]
        elif client_name == "json":
            return database._get_driver_database() / collection_name
        elif client_name == "migo":
            return database._get_driver_database()[collection_name]

    @staticmethod
    def _get_system_collection(driver_collection):
        client_name = RedB.get_client_name()

        if client_name == "mongo":
            try:
                from redb.mongo_system import MongoCollection
            except ImportError:
                raise ImportError(IMPORT_ERROR_MSG.format("mongo_system", "mongo"))

            return MongoCollection(driver_collection)

        elif client_name == "json":
            try:
                from redb.json_system import JSONCollection
            except ImportError:
                raise ImportError(IMPORT_ERROR_MSG.format("json_system", "json"))

            return JSONCollection(driver_collection)

        elif client_name == "migo":
            try:
                from redb.migo_system import MigoCollection
            except ImportError:
                raise ImportError(IMPORT_ERROR_MSG.format("migo_system", "migo"))

            return MigoCollection(driver_collection)

        raise ValueError(f"Unknown client: {client_name}")

    @staticmethod
    def _get_collection(instance_or_class: Type["BaseDocument"] | "BaseDocument"):
        driver_collection = BaseDocument._get_driver_collection(instance_or_class)
        system_collection = BaseDocument._get_system_collection(driver_collection)
        return system_collection

    @classmethod
    def collection_name(cls: Type["BaseDocument"]) -> str:
        return cls.__name__.lower()

    @classmethod
    def get_indexes(cls) -> list[Index | CompoundIndex]:
        return []

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        return [
            ClassField(model_field=field, base_class=cls)
            for field in cls.__fields__.values()
        ]

    @staticmethod
    def hash_function(string: str) -> str:
        return hashlib.sha3_256(string.encode("utf-8")).hexdigest()

    def _get_key_value_tuples_for_hash(
        self,
        fields: list[ClassField] | list[str],
        data: Dict[str, Any] | None = None,
    ) -> list[tuple[str, Any]]:
        if data:
            if isinstance(fields[0], str):
                transform = lambda field: (field, data[field])
            else:
                transform = lambda field: (field.join_attrs(), field.resolve(data))
        else:
            transform = lambda field: (field.join_attrs(), field.resolve(self))

        key_val_tuples = [transform(field) for field in fields]
        return key_val_tuples

    @staticmethod
    def _assemble_hash_string(fields: list[tuple[str, Any]]) -> str:
        return "|".join([str(val) for _, val in fields])

    def get_hash(
        self,
        data: Dict[str, Any] | None = None,
        use_data_fields: bool = False,
    ) -> str:
        if use_data_fields:
            fields = list(data.keys())
        else:
            fields = self.get_hashable_fields()

        if not fields:
            raise ValueError("No hashable fields found.")
        key_val_tuples = self._get_key_value_tuples_for_hash(fields, data)
        string = self._assemble_hash_string(key_val_tuples)
        return self.hash_function(string)

    def update_kwargs(self, kwargs):
        for field in self.__fields__.values():
            if field.alias not in kwargs:
                if field.required:
                    raise ValueError(
                        f"{field.alias} is missing for document {self.__class__.__name__}"
                    )
                kwargs[field.alias] = field.get_default()

        return kwargs

    def __repr__(self) -> str:
        class_name = self.__class__.__name__

        # add all pydantic attributes like attr=val, attr2=val2
        attributes = ", ".join(
            f"{field}={getattr(self, field)}" for field in self.__fields__
        )

        return f"{class_name}({attributes})"


def _apply_encoders(obj, encoders):
    obj_type = type(obj)
    if obj_type == list:
        obj = [_apply_encoders(val, encoders) for val in obj]
    elif obj_type == set:
        obj = {_apply_encoders(val, encoders) for val in obj}
    elif obj_type == tuple:
        obj = (_apply_encoders(val, encoders) for val in obj)
    elif obj_type == dict:
        obj = {
            _apply_encoders(key, encoders): _apply_encoders(val, encoders)
            for key, val in obj.items()
        }
    elif obj_type in encoders:
        encoding = encoders[obj_type]
        obj = encoding(obj) if callable(encoding) else encoding
    elif dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)
    else:
        for encoder_from, encoder_fn in encoders.items():
            if isinstance(obj, encoder_from):
                return encoder_fn(obj)
    return obj



from typing import Type

import pytest

from redb.core import BaseDocument, ClassField, Document


class Cat(BaseDocument):
    name: str


class Kitten(Document):
    is_bad_kitten: bool
    mom: Cat
    name: str

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        return [cls.name, cls.mom.name]


def test_key_val_tuples_construction():
    obj = Kitten(name="Fluffy", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    hashable_fields = obj.get_hashable_fields()
    key_val_tuples = obj._get_key_value_tuples_for_hash(hashable_fields)
    assert key_val_tuples == [("name", "Fluffy"), ("mom.name", "Whiskers")]


def test_hashable_fields_construction_order():
    class Doc(Document):
        attr1: str
        attr2: str
        attr3: str

        @classmethod
        def get_hashable_fields(cls) -> list[ClassField]:
            return [cls.attr1, cls.attr3]

    d = Doc(attr1="field1", attr2="field2", attr3="field3")
    hashable_fields = d.get_hashable_fields()
    key_val_tuples = d._get_key_value_tuples_for_hash(hashable_fields)
    assert key_val_tuples == [("attr1", "field1"), ("attr3", "field3")]

    class Doc(Document):
        attr1: str
        attr2: str
        attr3: str

        @classmethod
        def get_hashable_fields(cls) -> list[ClassField]:
            return [cls.attr2, cls.attr1]

    d = Doc(attr1="field1", attr2="field2", attr3="field3")
    hashable_fields = d.get_hashable_fields()
    key_val_tuples = d._get_key_value_tuples_for_hash(hashable_fields)
    assert key_val_tuples == [("attr2", "field2"), ("attr1", "field1")]


def test_equal_objects_equal_hash():
    obj1 = Kitten(name="Fluffy", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    obj1_hash = obj1.get_hash()
    obj2 = Kitten(name="Fluffy", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    obj2_hash = obj2.get_hash()
    assert obj1_hash == obj2_hash


def test_objects_only_different_by_non_hashable_fields():
    obj1 = Kitten(name="Fluffy", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    obj1_hash = obj1.get_hash()
    obj2 = Kitten(name="Fluffy", is_bad_kitten=True, mom=Cat(name="Whiskers"))
    obj2_hash = obj2.get_hash()
    assert obj1_hash == obj2_hash


def test_different_objects_different_hash():
    obj1 = Kitten(name="Fluffy", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    obj1_hash = obj1.get_hash()
    obj2 = Kitten(name="Oscar", is_bad_kitten=False, mom=Cat(name="Whiskers"))
    obj2_hash = obj2.get_hash()
    assert obj1_hash != obj2_hash



class REDBError(Exception):
    """Base class for all REDB errors."""

    def __init__(self, *args: object, collection_name: str = "") -> None:
        super().__init__(*args)
        self.collection_name = collection_name

    pass


class DocumentNotFound(REDBError):
    pass


class CannotUpdateIdentifyingField(REDBError):
    pass


class UniqueConstraintViolation(REDBError):
    def __init__(
        self, *args: object, dup_keys: dict, collection_name: str = ""
    ) -> None:
        if collection_name:
            msg = f"Duplicate key at collection {collection_name} on: {dup_keys}"
        else:
            msg = f"Duplicate key error on: {dup_keys}"
        super().__init__(msg, *args, collection_name=collection_name)
        self.dup_keys = [dup_keys]



from datetime import datetime
from typing import Any, Dict, Optional, TypeVar, Type

import pytz
from pymongo.collection import Collection as PyMongoCollection
from pymongo.errors import DuplicateKeyError

from ..core import RedB
from ..core.document import (
    Document,
    DocumentData,
    IncludeColumns,
    OptionalDocumentData,
    SortColumns,
    _format_document_data,
    _format_fields,
    _format_sort,
    _get_return_cls,
    _validate_fields,
)
from redb.interface.results import DeleteManyResult, ReplaceOneResult
from ..interface.errors import DocumentNotFound, UniqueConstraintViolation
from ..interface.fields import ClassField, Direction, Field, SortColumn
from ..interface.results import DeleteOneResult, InsertOneResult, UpdateOneResult

T = TypeVar("T", bound="IRememberDoc")
DOCUMENT_FIELDS = {"id", "created_at", "created_by"}
HISTORY_FIELDS = {"ref_id", "version", "retired_by", "retired_at"}


class IRememberDoc(Document):
    ref_id: str = ""
    version: int = 0
    retired_at: Optional[datetime] = None

    def __init__(self, **data):
        calculate_hash = False
        if "id" in data:
            data["_id"] = data.pop("id")
        if "_id" not in data:
            data["_id"] = None
            calculate_hash = True

        data = self.update_kwargs(data)
        if calculate_hash:
            data["_id"] = self.get_hash(data)

        data["ref_id"] = data["_id"]
        super().__init__(**data)

    @classmethod
    def _get_history_collection_driver(cls) -> PyMongoCollection:
        collection_name = cls.history_collection_name()
        database_name = cls.__database_name__

        client = RedB.get_client()
        database = (
            client.get_database(database_name)
            if database_name
            else client.get_default_database()
        )
        return database._get_driver_database()[collection_name]  # type: ignore

    @classmethod
    def _get_history_collection(cls):
        from redb.mongo_system import MongoCollection

        driver_collection = cls._get_history_collection_driver()
        return MongoCollection(driver_collection)

    @classmethod
    def _historical_insert_one(
        cls,
        data: DocumentData,
    ) -> InsertOneResult:
        _validate_fields(cls, data)

        collection = cls._get_history_collection()
        data = _format_document_data(data)
        try:
            return collection.insert_one(
                cls=cls,
                data=data,
            )
        except DuplicateKeyError as e:
            raise UniqueConstraintViolation(
                dup_keys=e.details["keyValue"],  # type: ignore
                collection_name=cls.history_collection_name(),
            )

    @classmethod
    def history_collection_name(cls):
        return f"{cls.collection_name()}-history"

    def dict(self, ignored_history_fields: bool = True, *args, **kwargs) -> dict:
        if ignored_history_fields:
            if "exclude" in kwargs and kwargs["exclude"] is not None:
                kwargs["exclude"] |= HISTORY_FIELDS
            else:
                kwargs["exclude"] = HISTORY_FIELDS

        out = super().dict(*args, **kwargs)
        return out

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        all_fields = super().get_hashable_fields()
        return list(filter(lambda x: x.model_field.name not in HISTORY_FIELDS, all_fields))

    @classmethod
    def historical_find_one(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        skip: int = 0,
    ) -> T:
        collection = cls._get_history_collection()
        filter = _format_document_data(filter)
        formatted_fields = _format_fields(fields)
        return_cls = _get_return_cls(cls, formatted_fields)
        return collection.find_one(
            cls=cls,
            return_cls=return_cls,  # type: ignore
            filter=filter,
            skip=skip,
            fields=formatted_fields,
        )

    @classmethod
    def historical_find_many(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        sort: SortColumns = SortColumn(name="version", direction=Direction.DESCENDING),
        skip: int = 0,
        limit: int = 0,
    ) -> list[T]:
        """
        Find many on the history collection.

        Allow users to query for previous revisions.
        Sort by version descending by default.
        """
        collection = cls._get_history_collection()
        filter = _format_document_data(filter)
        formatted_fields = _format_fields(fields)
        return_cls = _get_return_cls(cls, formatted_fields)
        sort_order = _format_sort(sort)
        return collection.find(
            cls=cls,
            return_cls=return_cls,  # type: ignore
            filter=filter,
            fields=formatted_fields,
            sort=sort_order,
            skip=skip,
            limit=limit,
        )  # type: ignore

    @classmethod
    def delete_history(
        cls,
        filter: OptionalDocumentData = None,
    ) -> DeleteManyResult:
        collection = cls._get_history_collection()
        filter = _format_document_data(filter)
        return collection.delete_many(
            cls=cls,
            filter=filter,
        )

    @classmethod
    def historical_update_one(
        cls,
        filter: DocumentData,
        update: DocumentData,
        upsert: bool = False,
        operator: str | None = "$set",
        allow_new_fields: bool = False,
        user_info: Any = None,
    ) -> UpdateOneResult:
        original_doc = super().find_one(filter=filter)
        new_history = cls.__build_history_from_ref(user_info, original_doc)
        update_result = cls.update_one(
            filter={"_id": original_doc.id},
            update=update,
            upsert=upsert,
            operator=operator,
            allow_new_fields=allow_new_fields,
        )
        cls._historical_insert_one(new_history)
        return update_result

    @classmethod
    def historical_replace_one(
        cls, filter: DocumentData, replacement: DocumentData, user_info: Any = None
    ) -> ReplaceOneResult:
        original_doc = super().find_one(filter=filter)
        new_history = cls.__build_history_from_ref(user_info, original_doc)
        replace_result = cls.replace_one(
            filter={"_id": original_doc.id},
            replacement=replacement,
        )
        cls._historical_insert_one(new_history)
        return replace_result


    @classmethod
    def historical_delete_one(
        cls,
        filter: DocumentData,
        user_info: Any = None,
    ) -> DeleteOneResult:
        original_doc = super().find_one(filter=filter)
        new_history = cls.__build_history_from_ref(user_info, original_doc)
        delete_result = cls.delete_one(filter={"_id": original_doc.id})
        cls._historical_insert_one(new_history)
        return delete_result

    @classmethod
    def __build_history_from_ref(
        cls,
        user_info: Any,
        referenced_doc: "IRememberDoc",
    ) -> Dict:
        history_filter = {"ref_id": referenced_doc.id}
        try:
            history = cls.historical_find_one(filter=history_filter, fields=["version"])
            version = history["version"] + 1  # type: ignore
        except DocumentNotFound:
            version = 1

        new_history = referenced_doc.dict(
            ignored_history_fields=False,
            exclude={"id"},
            exclude_none=True,
        )
        new_history["version"] = version
        new_history["retired_by"] = user_info.dict() if hasattr(user_info, "dict") else str(user_info)
        new_history["retired_at"] = pytz.UTC.localize(datetime.utcnow()).isoformat()
        new_history["ref_id"] = referenced_doc.id
        new_history["_id"] = f"{referenced_doc.id}_v{version}"
        return new_history



from abc import ABC, abstractmethod
from typing import Any, Type, TypeAlias

from redb.core import BaseDocument

from .fields import CompoundIndex, PyMongoOperations
from .results import (
    BulkWriteResult,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)

Json: TypeAlias = dict[str, Any]
OptionalJson: TypeAlias = dict[str, Any] | None
ReturnType: TypeAlias = BaseDocument | dict


class Collection(ABC):
    @abstractmethod
    def _get_driver_collection(self) -> "Collection":
        pass

    @abstractmethod
    def create_index(
        self,
        index: CompoundIndex,
    ) -> bool:
        pass

    @abstractmethod
    def find(
        self,
        cls: Type[BaseDocument],
        return_cls: Type[ReturnType],
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        sort: list[tuple[str, str | int]] | None = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[ReturnType]:
        pass

    @abstractmethod
    def find_one(
        self,
        cls: Type[BaseDocument],
        return_cls: Type[ReturnType],
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        skip: int = 0,
    ) -> ReturnType:
        pass

    @abstractmethod
    def distinct(
        self,
        cls: ReturnType,
        key: str,
        filter: OptionalJson = None,
    ) -> list[Any]:
        pass

    @abstractmethod
    def count_documents(
        self,
        cls: Type[BaseDocument],
        filter: OptionalJson = None,
    ) -> int:
        pass

    @abstractmethod
    def bulk_write(
        self,
        operations: list[PyMongoOperations],
    ) -> BulkWriteResult:
        pass

    @abstractmethod
    def insert_one(
        self,
        cls: Type[BaseDocument],
        data: Json,
    ) -> InsertOneResult:
        pass

    @abstractmethod
    def insert_many(
        self,
        cls: Type[BaseDocument],
        data: list[Json],
    ) -> InsertManyResult:
        pass

    @abstractmethod
    def replace_one(
        self,
        cls: Type[BaseDocument],
        filter: Json,
        replacement: Json,
        upsert: bool = False,
    ) -> ReplaceOneResult:
        pass

    @abstractmethod
    def update_one(
        self,
        cls: Type[BaseDocument],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateOneResult:
        pass

    @abstractmethod
    def update_many(
        self,
        cls: Type[BaseDocument],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateManyResult:
        pass

    @abstractmethod
    def delete_one(
        self,
        cls: Type[BaseDocument],
        filter: Json,
    ) -> DeleteOneResult:
        pass

    @abstractmethod
    def delete_many(
        self,
        cls: Type[BaseDocument],
        filter: Json,
    ) -> DeleteManyResult:
        pass



from functools import lru_cache
from typing import Any, Type, TypeVar

from migo.collection import BatchDocument
from migo.collection import Collection as MigoDriverCollection
from migo.collection import Document as MigoDocument
from migo.collection import Field as MigoField
from migo.collection import Filter as MigoFilter
from migo.utils import (
    AnnoyIndex,
    BINFlatIndex,
    BINIVFIndex,
    DISKANNIndex,
    FlatIndex,
    HNSWINdex,
)
from migo.utils import Index as MigoIndex
from migo.utils import (
    IVFFlatIndex,
    IVFPQIndex,
    IVFSQ8Index,
    MilvusBinaryIndex,
    MilvusFloatingIndex,
    MongoGeoIndex,
    MongoIndex,
)

from redb.core import Document
from redb.interface.collection import Collection, Json, OptionalJson, ReturnType
from redb.interface.fields import (
    ClassField,
    CompoundIndex,
    Direction,
    PyMongoOperations,
)
from redb.interface.results import (
    BulkWriteResult,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)

T = TypeVar("T", bound=Type[MigoDocument | MigoFilter])


def _is_milvus_field(field: ClassField) -> bool:
    return (
        hasattr(field.model_field, "vector_type")
        and field.model_field.vector_type is not None
    )


def _build_milvus_index_type(params: dict):
    if params["index_type"] == "FLAT":
        index_type = FlatIndex()
    elif params["index_type"] == "IVF_FLAT":
        index_type = IVFFlatIndex(nlist=params["nlist"])
    elif params["index_type"] == "IVF_SQ8":
        index_type = IVFSQ8Index(nlist=params["nlist"])
    elif params["index_type"] == "IVF_PQ":
        index_type = IVFPQIndex(
            nlist=params["nlist"],
            m=params["m"],
            nbits=params["nbits"],
        )
    elif params["index_type"] == "HNSW":
        index_type = HNSWINdex(
            M=params["M"],
            efConstruction=params["efCOnstruction"],
        )
    elif params["index_type"] == "ANNOY":
        index_type = AnnoyIndex(n_trees=params["n_trees"])
    elif params["index_type"] == "DISKANN*":
        index_type = DISKANNIndex()
    elif params["index_type"] == "BIN_FLAT":
        index_type = BINFlatIndex
    elif params["index_type"] == "BIN_IVF_FLAT":
        index_type = BINIVFIndex(nlist=params["nlist"])
    else:
        raise ValueError(f'Unknown index type: {params["index_type"]}')

    return index_type


def _build_index_name(index: CompoundIndex, fields: list[ClassField]) -> str:
    if index.name is None:
        index.name = "_".join([field.join_attrs("_") for field in fields])
        index.name = f"unique_{index.name}" if index.unique else index.name
        index.name = f"{index.direction.name.lower()}_{index.name}"

    return index.name


@lru_cache(maxsize=128)
def _build_migo_data(
    cls: Type[Document],
    data: OptionalJson,
    out: T,
) -> T | None:
    if data is None:
        return None

    mongo_data = {}
    milvus_data = {}
    for name, field in cls.__fields__.items():
        if name in data:
            if _is_milvus_field(field):
                milvus_data[name] = data[name]
            else:
                mongo_data[name] = data[name]

    if not mongo_data:
        mongo_data = None

    if not milvus_data:
        milvus_data = None

    return out(mongo_data, milvus_data)


@lru_cache(maxsize=128)
def _build_migo_fields(
    cls: Type[Document],
    fields: dict[str, bool] | None,
) -> list[MigoField] | None:
    if fields is None:
        return None

    migo_fields = []
    for name, field in cls.__fields__.items():
        if name in fields:
            if _is_milvus_field(field):
                migo_fields.append(MigoField(milvus_field=name))
            else:
                migo_fields.append(MigoField(mongo_field=name))

    return migo_fields


def _build_mongo_index(index: CompoundIndex) -> MongoIndex | MongoGeoIndex | None:
    mongo_fields: list[ClassField] = []
    for field in index.fields:
        if not _is_milvus_field(field):
            mongo_fields.append(field)

    if not mongo_fields:
        return None

    name = _build_index_name(index, mongo_fields)
    mongo_index = dict(
        key=[(field.join_attrs(), index.direction) for field in mongo_fields],
        name=name,
        unique=index.unique,
        type=index.direction,
        sparse=index.extras.get("sparse", False),
        expiration_secs=index.extras.get("expiration_secs", None),
        hidden=index.extras.get("hidden", None),
    )

    if index.direction in {Direction.GEO2D, Direction.GEOSPHERE}:
        mongo_index = MongoGeoIndex(
            bucket_size=index.extras.get("bucket_size", None),
            min=index.extras.get("min", None),
            max=index.extras.get("max", None),
            **mongo_index,
        )
    else:
        mongo_index = MongoIndex(**mongo_index)

    return mongo_index


def _build_milvus_indexes(
    index: CompoundIndex,
) -> list[MilvusBinaryIndex | MilvusFloatingIndex]:
    milvus_fields: list[ClassField] = []
    for field in index.fields:
        if _is_milvus_field(field):
            milvus_fields.append(field)

    if not milvus_fields:
        return []

    milvus_indexes = []
    for milvus_field in milvus_fields:
        name = _build_index_name(index, milvus_fields)
        index_type = _build_milvus_index_type(index.extras)
        milvus_index = dict(
            key=milvus_field.join_attrs(),
            name=name,
            metric_type=index.extras["metric_type"],
            index_type=index_type,
        )
        if type(index_type) in {BINFlatIndex, BINIVFIndex}:
            milvus_indexes.append(MilvusBinaryIndex(**milvus_index))
        else:
            milvus_indexes.append(MilvusFloatingIndex(**milvus_index))

    return milvus_indexes


class MigoCollection(Collection):
    __client_name__: str = "migo"

    def __init__(self, collection: MigoDriverCollection) -> None:
        super().__init__()

        self.__collection = collection

    def _get_driver_collection(self) -> MigoDriverCollection:
        return self.__collection

    def create_index(
        self,
        index: CompoundIndex,
    ) -> bool:
        mongo_index = _build_mongo_index(index)
        milvus_indexes = _build_milvus_indexes(index)
        try:
            self.__collection.create_indexes(
                [MigoIndex(mongo_index=mongo_index, milvus_indexes=milvus_indexes)]
            )
            return True
        except:
            return False

    def find(
        self,
        cls: Type[Document],
        return_cls: ReturnType,
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        sort: list[tuple[str, str | int]] | None = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[ReturnType]:
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        migo_fields = _build_migo_fields(cls, fields)
        results = self.__collection.find_many(
            filter=migo_filter,
            fields=migo_fields,
            sort=sort,
            limit=limit,
        )
        return [return_cls(**result) for result in results]

    def find_one(
        self,
        cls: Type[Document],
        return_cls: ReturnType,
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        skip: int = 0,
    ) -> ReturnType:
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        migo_fields = _build_migo_fields(cls, fields)
        result = self.__collection.find_one(filter=migo_filter, fields=migo_fields)
        return return_cls(**result)

    def distinct(
        self,
        cls: ReturnType,
        key: str,
        filter: OptionalJson = None,
    ) -> list[Any]:
        results = self.__collection.distinct(key=key, filter=filter)
        return results

    def count_documents(
        self,
        cls: Type[Document],
        filter: OptionalJson = None,
    ) -> int:
        return self.__collection.count(filter=filter)

    def bulk_write(self, _: list[PyMongoOperations]) -> BulkWriteResult:
        raise NotImplementedError

    def insert_one(
        self,
        cls: Type[Document],
        data: Json,
    ) -> InsertOneResult:
        migo_data = _build_migo_data(cls, data=data, out=MigoDocument)
        result = self.__collection.insert_one(data=migo_data)
        return InsertOneResult(inserted_id=result.inserted_id)

    def insert_many(
        self,
        cls: Type[Document],
        data: list[Json],
    ) -> InsertManyResult:
        migo_data = BatchDocument(mongo_documents=[], milvus_arrays=[])
        for value in data:
            migo_doc = _build_migo_data(cls, data=value, out=MigoDocument)
            if migo_doc.mongo_document is not None:
                migo_data.mongo_documents.append(migo_doc.mongo_document)
            if migo_doc.milvus_array:
                migo_data.milvus_arrays.append(migo_doc.milvus_array)

        result = self.__collection.insert_many(migo_data)
        return InsertManyResult(inserted_ids=result.inserted_ids)

    def replace_one(
        self,
        cls: Type[Document],
        filter: Json,
        replacement: Json,
        upsert: bool = False,
    ) -> ReplaceOneResult:
        migo_doc = _build_migo_data(cls, data=replacement, out=MigoDocument)
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        result = self.__collection.replace_one(
            data=migo_doc,
            filter=migo_filter,
            upsert=upsert,
        )
        return ReplaceOneResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def update_one(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateOneResult:
        migo_doc = _build_migo_data(cls, data=update, out=MigoDocument)
        migo_filter = _build_migo_data(cls, data=filter, out=MigoDocument)
        result = self.__collection.update_one(
            data=migo_doc,
            filter=migo_filter,
            upsert=upsert,
        )
        return UpdateOneResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def update_many(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateManyResult:
        migo_doc = _build_migo_data(cls, data=update, out=MigoDocument)
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        result = self.__collection.update_many(
            data=migo_doc,
            filter=migo_filter,
            upsert=upsert,
        )
        return UpdateManyResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def delete_one(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteOneResult:
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        result = self.__collection.delete_one(filter=migo_filter)
        return DeleteOneResult(deleted_count=result.deleted_count)

    def delete_many(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteManyResult:
        migo_filter = _build_migo_data(cls, data=filter, out=MigoFilter)
        result = self.__collection.delete_many(filter=migo_filter)
        return DeleteManyResult(deleted_count=result.deleted_count)



from __future__ import annotations

import shlex
import subprocess
from pathlib import Path

import setuptools


def read_multiline_as_list(file_path: Path | str) -> list[str]:
    with open(file_path) as req_file:
        contents = req_file.read().split("\n")
        if contents[-1] == "":
            contents.pop()
        return contents


def get_optional_requirements() -> dict[str, list[str]]:
    """Get dict of suffix -> list of requirements."""
    requirements_files = Path(".").glob(r"requirements-*.txt")
    requirements = {
        p.stem.split("-")[-1]: read_multiline_as_list(p) for p in requirements_files
    }
    return requirements


def get_version() -> str:
    raw_git_cmd = "git describe --tags"
    git_cmd = shlex.split(raw_git_cmd)
    fmt_cmd = shlex.split("cut -d '-' -f 1,2")
    git = subprocess.Popen(git_cmd, stdout=subprocess.PIPE)
    cut = subprocess.check_output(fmt_cmd, stdin=git.stdout)
    ret_code = git.wait()
    assert ret_code == 0, f"{raw_git_cmd!r} failed with exit code {ret_code}."
    return cut.decode().strip()


requirements = read_multiline_as_list("requirements.txt")

opt_requirements = get_optional_requirements()
opt_requirements["all"] = [
    value
    for key, values in opt_requirements.items()
    if key != "systems"
    for value in values
]

with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name="redb-odm",
    version=get_version(),
    author="Teia Labs",
    author_email="contato@teialabs.com",
    description="A python ODM for JSON, Mongo, and Mongo+Milvus.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/teialabs/redb",
    packages=setuptools.find_namespace_packages(exclude=["tests"]),
    keywords="database milvus mongo json interface",
    python_requires=">=3.10",
    install_requires=requirements,
    extras_require=opt_requirements,
)



from typing import Type, TypeVar

from ..core import Document
from ..core.document import IncludeColumns, OptionalDocumentData, SortColumns
from ..interface.errors import DocumentNotFound
from ..interface.fields import ClassField, IncludeColumn

T = TypeVar("T")


class SoftDeletinDoc(Document):
    is_deleted: bool = False

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        all_fields = super().get_hashable_fields()
        return list(filter(lambda x: x.model_field.name != "is_deleted", all_fields))

    @classmethod
    def soft_delete_one(cls, filters):
        result = cls.update_one(filters, dict(is_deleted=True))
        if not result.matched_count:
            raise DocumentNotFound

    @classmethod
    def soft_undelete_one(cls, filters):
        result = cls.update_one(filters, dict(is_deleted=False))
        if not result.matched_count:
            raise DocumentNotFound

    @classmethod
    def soft_delete_many(cls, filters):
        result = cls.update_many(filters, dict(is_deleted=True))
        if not result.matched_count:
            raise DocumentNotFound

    @classmethod
    def soft_undelete_many(cls, filters):
        result = cls.update_many(filters, dict(is_deleted=False))
        if not result.matched_count:
            raise DocumentNotFound

    @classmethod
    def find_one(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        skip: int = 0,
    ) -> T:
        if filter is None:
            filters = {}
        elif isinstance(filter, dict) and "is_deleted" not in filter:
            filters = filter | {"is_deleted": False}
        else:
            filters = filter
        # if it were an instance of Document, it would either
        # already be using the default value
        # or the user would have set his own. Either way, it's ok.
        if fields is None:
            fields = [IncludeColumn(name="is_deleted", include=False)]
        # if it's a list of strings or a list of include columns, it's fine.
        return super().find_one(filter=filters, fields=fields, skip=skip)

    @classmethod
    def find_many(
        cls: Type[T],
        filter: OptionalDocumentData = None,
        fields: IncludeColumns = None,
        sort: SortColumns = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[T]:
        if filter is None:
            filters = {}
        elif isinstance(filter, dict) and "is_deleted" not in filter:
            filters = filter | {"is_deleted": False}
        else:
            filters = filter
        # if it were an instance of Document, it would either
        # already be using the default value
        # or the user would have set his own. Either way, it's ok.
        if fields is None:
            fields = [IncludeColumn(name="is_deleted", include=False)]
        # if it's a list of strings, it's fine.
        # TODO: if it's a list of include columns, we should check whether
        # it's removals only and add the is_deleted removal.
        result = super().find_many(
            filter=filters, fields=fields, sort=sort, skip=skip, limit=limit
        )
        return result  # type: ignore



import os
from typing import Optional

import pytest
from pymongo import MongoClient
from pymongo.database import Database

from redb.behaviors import IRememberDoc
from redb.core import RedB
from redb.interface.configs import MongoConfig
from redb.interface.errors import DocumentNotFound
from redb.interface.fields import ClassField


class Cat(IRememberDoc):
    name: str
    breed: str = "Domestic Shorthair"
    created_by: str
    retired_by: Optional[str] = None

    @classmethod
    def get_hashable_fields(cls) -> list[ClassField]:
        return [cls.name]  # type: ignore

    @classmethod
    def collection_name(cls) -> str:
        return "cats"


@pytest.fixture(scope="module", autouse=True)
def client():
    RedB.setup(
        MongoConfig(
            database_uri=os.environ["MONGODB_URI"],
        )
    )


@pytest.fixture(scope="module")
def db() -> Database:
    return MongoClient(os.environ["MONGODB_URI"]).get_default_database()


@pytest.fixture(scope="session")
def creator_email():
    return "creator@cats.com"


@pytest.fixture(scope="function")
def fluffy_cat(creator_email) -> Cat:
    return Cat(name="Fluffy", created_by=creator_email)


@pytest.fixture(scope="function")
def pony_cat(creator_email) -> Cat:
    return Cat(name="Pony", created_by=creator_email)


@pytest.fixture(scope="function")
def little_cat(creator_email) -> Cat:
    return Cat(name="Little", created_by=creator_email)


@pytest.fixture(scope="function", autouse=True)
def teardown(fluffy_cat: Cat, pony_cat: Cat, little_cat: Cat):
    yield
    cat_names = [fluffy_cat.name, pony_cat.name, little_cat.name]
    filters = {"name": {"$in": cat_names}}
    Cat.delete_many(filters)
    Cat.delete_history(filters)


@pytest.fixture(scope="module")
def user_email():
    return "test_user@mail.com"


def test_historical_delete_one(db: Database, fluffy_cat: Cat, creator_email: str, user_email: str):
    obj = fluffy_cat
    obj.insert()
    deleted_result = Cat.historical_delete_one({"_id": obj.id}, user_info=user_email)
    assert deleted_result.deleted_count
    cats = list(db["cats"].find({"name": fluffy_cat.name}))
    assert len(cats) == 0
    hist_cats = list(db["cats-history"].find({"name": fluffy_cat.name}))
    assert len(hist_cats) == 1
    hist_cat = hist_cats[0]
    assert hist_cat["version"] == 1
    assert hist_cat["created_by"] == creator_email
    assert hist_cat["retired_by"] == user_email


@pytest.mark.order(after="test_historical_delete_one")
def test_historical_find_one(user_email, pony_cat: Cat):
    pony_cat.insert()
    del_res = Cat.historical_delete_one({"_id": pony_cat.id}, user_info=user_email)
    assert del_res.deleted_count
    history = Cat.historical_find_one(filter={"name": pony_cat.name})
    assert history.version == 1
    assert history.name == pony_cat.name


@pytest.mark.order(after="test_historical_delete_one")
def test_historical_find_many(little_cat: Cat):
    little_cat_obj = little_cat.dict(exclude={"created_at"})
    Cat(**little_cat_obj).insert()

    del_res = Cat.historical_delete_one({"_id": little_cat.id})
    assert del_res.deleted_count

    Cat(**little_cat_obj).insert()

    del_res = Cat.historical_delete_one({"_id": little_cat.id})
    assert del_res.deleted_count
    histories = Cat.historical_find_many({"name": little_cat.name})
    assert len(histories) == 2
    assert histories[0].version == 2
    assert histories[1].version == 1
    latest = histories[0]
    previous = histories[1]
    assert latest.name == previous.name
    assert latest.created_at >= previous.created_at
    assert ((l := latest.retired_at) is not None) and ((p := previous.retired_at) is not None) and l > p


def test_historical_update_one(db: Database, creator_email: str, user_email: str, fluffy_cat: Cat):
    fluffy_cat.insert()
    updated_result = Cat.historical_update_one(
        {"name": fluffy_cat.name}, {"breed": "American Bobtail"}, user_info=user_email
    )
    assert updated_result.modified_count
    hist_cats = list(db["cats-history"].find({"name": fluffy_cat.name}))
    assert len(hist_cats) == 1
    hist_cat = hist_cats[0]
    assert hist_cat["version"] == 1
    assert hist_cat["created_by"] == creator_email
    assert hist_cat["retired_by"] == user_email
    assert hist_cat["breed"] == fluffy_cat.breed
    cats = list(db["cats"].find({"name": fluffy_cat.name}))
    assert len(cats) == 1
    cat = cats[0]
    assert cat["breed"] == "American Bobtail"


def test_historical_replace_one(db: Database, creator_email: str, user_email: str, fluffy_cat: Cat):
    fluffy_cat.insert()
    old_cat_id = fluffy_cat.id
    replaced_result = Cat.historical_replace_one(
        {"name": fluffy_cat.name}, {**fluffy_cat.dict(), "breed": "American Bobtail"}, user_info=user_email
    )
    assert replaced_result.modified_count
    hist_cats = list(db["cats-history"].find({"name": fluffy_cat.name}))
    assert len(hist_cats) == 1
    hist_cat = hist_cats[0]
    assert hist_cat["version"] == 1
    assert hist_cat["ref_id"] == old_cat_id
    assert hist_cat["created_by"] == creator_email
    assert hist_cat["retired_by"] == user_email
    assert hist_cat["breed"] == fluffy_cat.breed
    cats = list(db["cats"].find({"name": fluffy_cat.name}))
    assert len(cats) == 1
    cat = cats[0]
    assert cat["breed"] == "American Bobtail"



from typing import Literal

from redb.interface.configs import (
    CONFIG_TYPE,
    JSONConfig,
    MigoConfig,
    MongoConfig,
    check_config,
)


class RedB:
    """Client singleton."""

    _client = None
    _client_name: str | None = None
    _config: CONFIG_TYPE | None = None

    @classmethod
    def get_client(cls):
        if cls._client is None:
            raise RuntimeError("Client not setup. Call setup() first.")
        return cls._client

    @classmethod
    def get_client_name(cls) -> str:
        if cls._client_name is None:
            raise RuntimeError("Client not setup. Call setup() first.")
        return cls._client_name
    
    @classmethod
    def get_config(cls) -> CONFIG_TYPE:
        if cls._config is None:
            raise RuntimeError("Client not setup. Call setup() first.")
        return cls._config

    @classmethod
    def setup(
        cls,
        config: CONFIG_TYPE,
        backend: Literal["json", "mongo"] | None = None,
    ) -> None:
        if backend is None and isinstance(config, dict):
            raise ValueError("Cannot determine client type from backend and config")
        elif backend == "json" or (
            backend is None and check_config(config, JSONConfig)
        ):
            from redb.json_system import JSONClient

            cls._client = JSONClient(config)
            cls._client_name = "json"

        elif backend == "mongo" or (
            backend is None and check_config(config, MongoConfig)
        ):
            from redb.mongo_system import MongoClient

            cls._client = MongoClient(config)
            cls._client_name = "mongo"

        elif backend == "migo" or (
            backend is None and check_config(config, MigoConfig)
        ):
            from redb.migo_system import MigoClient

            cls._client = MigoClient(config)
            cls._client_name = "migo"

        else:
            raise ValueError(f"Backend not found for config type: {type(config)!r}.")
        
        cls._config = config



import json
from pathlib import Path

import pytest

from .utils import Embedding, read_json, remove_document


@pytest.fixture
def embedding():
    return Embedding(
        kb_name="KB",
        model="ai",
        text="Some data.",
        vector=[1, 2],
        source_url="www",
    )


def test_insert_one(json_client, collection_path: Path, embedding: Embedding):
    # Ignored while we do not have a Mongo instance to run agains
    # with transaction(collection_class=Embedding, config=MongoConfig(database_uri="mongodb://localhost:27017/", default_database="teia"), database_name="another") as embedding:
    #     embedding.insert_one(data=d)
    #     assert not (collection_path / f"{d.id}.json").is_file()
    Embedding.insert_one(embedding)
    assert (collection_path / f"{embedding.id}.json").is_file()
    json_path = collection_path / f"{embedding.id}.json"
    other = read_json(json_path)
    remove_document(collection_path, embedding.id)
    assert other == embedding


def test_find_by_id(json_client, collection_path: Path, embedding: Embedding):
    Embedding.insert_one(embedding)
    found_instance = Embedding.find_one(filter=dict(_id=embedding.id))
    remove_document(collection_path, embedding.id)
    assert found_instance == embedding


def test_insert_vectors(json_client, collection_path: Path):
    data = dict(
        _id=["a", "b", "c"],
        kb_name=["KB5", "KB6", "KB7"],
        model=["ai", "ai", "ai"],
        text=["Some data 5", "Some data 6", "Some data 7"],
        vector=[[1, 3], [2, 4], [3, 5]],
        source_url=["ww5", "ww6", "ww7"],
    )
    Embedding.insert_vectors(data)
    embeddings = [Embedding(**read_json(val)) for val in collection_path.glob("*.json")]
    assert sorted(e.id for e in embeddings) == data["_id"]
    embeddings.sort(key=lambda embedding: embedding.kb_name)
    for i in range(len(embeddings)):
        original = Embedding(
            id=data["_id"][i],
            kb_name=data["kb_name"][i],
            model=data["model"][i],
            text=data["text"][i],
            vector=data["vector"][i],
            source_url=data["source_url"][i],
        )
        other = embeddings[i]
        remove_document(collection_path, other.id)
        # Make both timestamps be the same because they are being generated on object instantiation
        original.created_at = other.created_at
        original.updated_at = other.updated_at
        assert original == other


def test_count_documents(json_client, collection_path: Path, embedding: Embedding):
    doc_count = Embedding.count_documents()
    assert doc_count == 0
    Embedding.insert_one(embedding)
    doc_count = Embedding.count_documents()
    try:
        assert doc_count == 1
    finally:
        remove_document(collection_path, embedding.id)


def test_replace_one(json_client, collection_path: Path, embedding: Embedding):
    Embedding.insert_one(embedding)
    replacement = Embedding(
        kb_name="KB_REPLACED",
        model="replaced",
        text="Some replaced data.",
        vector=[114, 101, 112, 108, 97, 99, 101, 100],
        source_url="www.replaced",
    )
    Embedding.replace_one(embedding, replacement)
    other = read_json(collection_path / f"{replacement.id}.json")
    remove_document(collection_path, replacement.id)
    assert replacement == other


def test_update_one(json_client, collection_path: Path, embedding: Embedding):
    Embedding.insert_one(embedding)
    Embedding.update_one(embedding, update={"vector": [1, 2, 3]})
    expected = embedding.dict()
    expected["vector"] = [1, 2, 3]
    other = read_json(collection_path / f"{embedding.id}.json")
    remove_document(collection_path, embedding.id)
    for key, value in expected.items():
        if key == "updated_at":
            continue

        assert key in other
        assert value == other[key]


def test_delete_one(json_client, collection_path: Path, embedding: Embedding):
    try:
        Embedding.insert_one(embedding)
    except:
        pass
    Embedding.delete_one(embedding)
    with pytest.raises(AssertionError):
        assert (collection_path / f"{embedding.id}.json").is_file()



import os

import pytest
from pymongo import MongoClient
from pymongo.database import Database

from redb.core import Document, RedB, MongoConfig
from redb.interface.errors import UniqueConstraintViolation
from redb.interface.fields import CompoundIndex, Index


@pytest.fixture(scope="module", autouse=True)
def client():
    RedB.setup(
        MongoConfig(
            database_uri=os.environ["MONGODB_URI"],
        )
    )

@pytest.fixture(scope="module", autouse=True)
def db():
    return MongoClient(os.environ["MONGODB_URI"]).get_database()

class Cat(Document):
    name: str
    breed: str = "Domestic Shorthair"
    created_by: str

    @classmethod
    def get_indexes(cls) -> list[Index | CompoundIndex]:
         return [CompoundIndex([cls.name, cls.breed], unique=True)]  # type: ignore

    @classmethod
    def collection_name(cls) -> str:
        return "cats"


def test_unique_constraints(db: Database):
    db["cats"].drop()
    Cat.create_indexes()
    cat = Cat(name="Kitty", created_by="me")
    result = Cat.insert_one(cat)
    assert result.inserted_id is not None
    assert cat == Cat.find_one({"_id": result.inserted_id})
    cat2 = cat = Cat(name="Kitty", created_by="other")
    with pytest.raises(UniqueConstraintViolation):
        Cat.insert_one(cat2)
    try:
        Cat.insert_one(cat2)
    except UniqueConstraintViolation as e:
        assert e.collection_name == "cats"
        assert e.dup_keys == [{"name": "Kitty", "breed": "Domestic Shorthair"}]



from typing import Any, Literal, Type

from pymongo.collection import Collection as PymongoCollection

from redb.core import Document
from redb.interface.collection import Collection, Json, OptionalJson, ReturnType
from redb.interface.fields import CompoundIndex, PyMongoOperations
from redb.interface.results import (
    BulkWriteResult,
    DeleteManyResult,
    DeleteOneResult,
    InsertManyResult,
    InsertOneResult,
    ReplaceOneResult,
    UpdateManyResult,
    UpdateOneResult,
)
from redb.interface.errors import DocumentNotFound


class MongoCollection(Collection):
    __client_name__: str = "mongo"

    def __init__(self, collection: PymongoCollection) -> None:
        super().__init__()

        self.__collection = collection

    def _get_driver_collection(self) -> PymongoCollection:
        return self.__collection

    def create_index(
        self,
        index: CompoundIndex,
    ) -> bool:
        name = index.name
        if name is None:
            name = "_".join([field.join_attrs("_") for field in index.fields])
            name = f"unique_{name}" if index.unique else name
            name = f"{index.direction.name.lower()}_{name}_index"
        try:
            keys = [
                (field.join_attrs(), index.direction.value) for field in index.fields
            ]
            self.__collection.create_index(
                keys=keys,
                name=name,
                unique=index.unique,
            )
            return True
        except:
            return False

    def find(
        self,
        cls: Type[Document],
        return_cls: ReturnType,
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        sort: list[tuple[str, str | int]] | None = None,
        skip: int = 0,
        limit: int = 0,
    ) -> list[ReturnType]:
        return [
            return_cls(**result)
            for result in self.__collection.find(
                filter=filter,
                projection=fields,
                sort=sort,
                skip=skip,
                limit=limit,
            )
        ]

    def find_one(
        self,
        cls: Type[Document],
        return_cls: ReturnType,
        filter: OptionalJson = None,
        fields: dict[str, bool] | None = None,
        skip: int = 0,
    ) -> ReturnType:
        result = self.__collection.find_one(
            filter=filter,
            projection=fields,
            skip=skip,
        )
        if not result:
            m = f"Document not found with filters {filter} in collection {self.__collection.name}."
            raise DocumentNotFound(m)
        return return_cls(**result)

    def distinct(
        self,
        cls: ReturnType,
        key: str,
        filter: OptionalJson = None,
    ) -> list[Any]:
        results = self.__collection.distinct(key=key, filter=filter)
        return results

    def count_documents(
        self,
        cls: Type[Document],
        filter: OptionalJson = None,
    ) -> int:
        return self.__collection.count_documents(filter=filter)

    def bulk_write(
        self,
        operations: list[PyMongoOperations],
    ) -> BulkWriteResult:
        result = self.__collection.bulk_write(requests=operations)
        return BulkWriteResult(
            deleted_count=result.deleted_count,
            inserted_count=result.inserted_count,
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_count=result.upserted_count,
            upserted_ids=result.upserted_ids,
        )

    def insert_one(
        self,
        cls: Type[Document],
        data: dict,
    ) -> InsertOneResult:
        result = self.__collection.insert_one(document=data)
        return InsertOneResult(inserted_id=result.inserted_id)

    def insert_many(
        self,
        cls: Type[Document],
        data: list[Json],
    ) -> InsertManyResult:
        result = self.__collection.insert_many(documents=data)
        return InsertManyResult(inserted_ids=result.inserted_ids)

    def replace_one(
        self,
        cls: Type[Document],
        filter: Json,
        replacement: Json,
        upsert: bool = False,
    ) -> ReplaceOneResult:
        result = self.__collection.replace_one(
            filter=filter,
            replacement=replacement,
            upsert=upsert,
        )
        return ReplaceOneResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def update_one(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateOneResult:
        result = self.__collection.update_one(
            filter=filter,
            update=update,
            upsert=upsert,
        )
        return UpdateOneResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def update_many(
        self,
        cls: Type[Document],
        filter: Json,
        update: Json,
        upsert: bool = False,
    ) -> UpdateManyResult:
        result = self.__collection.update_many(
            filter=filter,
            update=update,
            upsert=upsert,
        )
        return UpdateManyResult(
            matched_count=result.matched_count,
            modified_count=result.modified_count,
            upserted_id=result.upserted_id,
        )

    def delete_one(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteOneResult:
        res = self.__collection.delete_one(filter=filter)
        return DeleteOneResult(deleted_count=res.deleted_count)

    def delete_many(
        self,
        cls: Type[Document],
        filter: Json,
    ) -> DeleteManyResult:
        result = self.__collection.delete_many(filter=filter)
        return DeleteManyResult(deleted_count=result.deleted_count)
